/* -*-Mode:C; tab-width:4; indent-tabs-mode:t; c-file-style:"stroustrup";-*- */
/* Main code for Project */
//#define ERROR_CHECK_LEVEL 2
#include <PalmOS.h>
#include <DateGlue.h>
#include <TxtGlue.h>
#include <FntGlue.h>
#include <TsmGlue.h>
#include <WinGlue.h>
#include "handera/Vga.h"
//#include "SonyCLIE.h"

#include "link.h"
#include "task.h"
#include "xb.h"
#include "progectdb.h"
#include "progectRsc.h"
#include "progect.h"
#include "flat.h"
#include "btm.h"
#include "beam.h"
#include "projlistMODEL.h"

// icon support
#include "icon.h"
// ToDo export support
#include "ToDoDB.h"
// Memo linking support
#include "MemoDB.h"
// address linking support
#include "AddressDB.h"
// Doc export support
#include "doc.h"
// Category support
#include "category.h"
// history support
#include "history.h"
// hotsync support
#include "aftersync.h"

#include "projlistGUI.h"

// not used yet
typedef enum {tblRedrawAll, tblRedrawTable, tblUpdateRedrawAll, 
	tblUpdateRedrawTable} redrawTableType;

// for FrmTaskEditPart
typedef enum {showValue, hideValue} PartType;

// prototypes

// events
void           ProgressHandleEvent(void) UISECT;
void           PriorityHandleEvent(EventPtr e) UISECT;
UInt16         SelectRange(void) UISECT;
Boolean        FrmMainMenuHandleEvent(EventPtr e) UISECT;
Boolean        FrmMainCtlSelectHandleEvent(EventPtr e) UISECT;
Boolean        FrmMainKeyDownHandleEvent(EventPtr e) UISECT;
static Boolean FrmMainHandleEvent (EventPtr e) UISECT;
static void    UpdateIconButton(void) UISECT;
static Boolean FrmTaskEditHandleEvent (EventPtr e) UISECT;
Boolean        HandleChooseNameButtonOK(void) UISECT;
static Boolean FrmMemoExportPropertiesHandleEvent (EventPtr e) UISECT;
Boolean        EditMenuHandleEvent(UInt16 menuid, FieldPtr fld) UISECT;
void           FrmNoteEditUpdateScroll(void) UISECT;
void           FrmNoteEditScrollPage(WinDirectionType direction) UISECT;
static Boolean FrmNoteEditHandleEvent (EventPtr e) UISECT;
Boolean        FrmFlatCtlSelectHandleEvent(EventPtr e) UISECT;
void           ProjectFlatTableUpdate(void) UISECT;
static Boolean FrmFlatHandleEvent (EventPtr e) UISECT;
static Boolean FrmFlatFilterHandleEvent (EventPtr e) UISECT;
static Boolean FrmDisplayPropertiesHandleEvent (EventPtr e) UISECT;
static Boolean FrmTaskDefaultsHandleEvent (EventPtr e) UISECT;
void           UpdatePopupTrigger(MemPtr trigger, MemPtr list) UISECT;
static void    FrmTaskEditPart(PartType part) UISECT;
static Boolean FrmProjectPropertiesHandleEvent (EventPtr e);

static Boolean ApplicationHandleEvent(EventPtr e);
static UInt16 StartApplication(void);
static void StopApplication(void);
static void EventLoop(void);
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags);
static void Search(FindParamsPtr findParams);

// icons
#define IconNum 16
MemHandle gIconHandle[IconNum]; // to hold handles to the bitmaps


// Globals

OSCaps_t OSCaps = {
	false, false, false, false, false, false, false, false,
	0, false, 0, 160, 160, 0
};


UInt16 gCardNo=0; // card number of open database
DmOpenRef gdbP = NULL; // actual database
DmOpenRef gClip = NULL; // clipboard database
UInt16 gActualTask = 0; // selected task
UInt16 gLastSelected = 0; // last selected task
UInt32 gFlatOffset = 0; // flat view offset of first item

TaskType gEmptyTask = {
	{ // attr
		{
			0, // level
			0, // next
			0, // child
			0, // opened
			0, // prev
			0  // reserved
		}
	}, 
	{ // format
		{
			0,            // hasStartDate
			0,            // hasPred
			0,            // hasDuration
			1,            // hasDueDate
			0,            // hasToDo
			0,            // hasNote
			0,            // hasLink
			progressType, // itemType
			0,            // hasXB
			1,            // newTask
			1,            // newFormat
			0             // nextFormat
		}
	},
	NO_PRIORITY, // priority
	0,           // completed
	{ // due date
		0, 
		0, 
		0
	}, 
	"", // desc
	""  // reserved (align)
};

DateType gNoDate = {0, 0, 0};
Char gDateStr[dowLongDateStrLength]; // to store a date string
UInt8 gLevelOffset = 0; // for horizontal scroll and subtree viewing
UInt16 gParentTask = 0; // the parent of all tasks for viewing purposes only
UInt8 gRefLevel = 0; // level of gParentTask;
Int16 gLastDownX, gLastDownY;
TablePtr gTable; // main form table
LocalID gdbID = 0; // id of opened database (set by OpenDB and CloseDB)
// used for date differences calculations
UInt8 gLimits[] = { 0, 1, 2, 3, 4, 5, 6, 7, 14, 21, 28 }; // UInt8 is to saving memory.
DateType gToday;
UInt32 gTodayDays; // to speed up day comparaisons calculations
Char* docName; // use by FrmChooseName
UInt8 ChooseNameFunction = ChooseNameCreate;
UInt16 gProjectTableMaxY = 0;
UInt8 gActualNumRows = 11; // number of rows to show
// 0.23
//UInt16 gNumDeleted = 0;   // nb of deleted records in actual database

// Preferences
DateFormatType gDateFormat; // actual date format (from palm, StartApplication)
ProjectPrefsType gProjectPrefs = 
{
	DBVERSION, // db format
	0, 		// reserved
	false,	// hideDoneTasks
	true, 	// displayDueDates
	true, 	// displayPriorities
	false,  // displayYear
	false,  // useFatherStatus
	true,	// autoSyncToDo
	true, 	// flatHideDone
	2, 		// flatDated (0 = no, 1 = yes, 2 = not used)
	3, 		// flatMinPriority
	true, 	// flatOr
	true,  	// flatMin
	0, 		// boldMinPriority
	0,      // boldMinDays
	false,  // strikeDoneTasks
	false,  // hideDoneProgress
	false,  // hideProgress
	// taskDefaults
	{{{0, 0, 0, 0, 0, 0}}, {{0, 0, 0, 1, 0, 0, 0, progressType, 0, 1, 1, 0}},
	NO_PRIORITY, 0, {0, 0, 0}, "", ""},
	noSort, // flatSorted
	0,      // flatDateLimit
	true,   // completionDate
	0xffff, // flatCategories
	2,      // wordWrapLines
	true    // drawTreeLines
};
CurrentPrefsType gCurrentPrefs; // current prefs, used to restore applic state
SavedPrefsType gSavedPrefs = {true}; // saved prefs, used to restore global 
                                     // application prefs
MemoExportOptionsType gMemoExportPrefs =
	{
		true,  // exportDone;
		true,  // exportProgress;
		true,  // exportDueDate;
		true,  // exportPriority;
		false, // exportNote;
		false, // exportOnlyTerminals;
		DOC_EXPORT_RAWTEXT // exportDocFormat;
	};


static UInt16 wVerticalLines[25];


/****************************************************************************
 * Name : TblGetRowContainingPoint
 * Desc : Return the table row that bounds the given x & y screen coordinates.
 *        Returns -1 if there is no usable row that contains the point.
 * Parm : 
 * 			-> The table to operate on
 *          -> x, y screen coordinates
 * Out  :   <- The row number containing (x, y), or -1 if none.
 * Auth : hcc, 05.05.2001
 ***************************************************************************/
Int16 TblGetRowContainingPoint(const TableType *table, UInt16 x, UInt16 y)
{
	RectangleType tableBounds;
	Int16 row, lastRow;
	UInt16 rowY, rowH;

	// Find top Y of the first row.
	TblGetBounds(table, &tableBounds);
	rowY = tableBounds.topLeft.y;

	for(row=0,lastRow = TblGetLastUsableRow(table); row <= lastRow; row++)
	{
		// NOTE: Initially tried just TblGetItemBounds for each row, but
		// the extent.y was always 11, even if the row height was several
		// lines long, so now just add up all the heights from the start.

		rowH = TblGetRowHeight(table, row);
		if(rowY <= y && y < rowY + rowH)
			break;
		rowY += rowH;
	}
	if(row > lastRow) // Not in any usable row
		row = -1;
	DBGMSG((DBB,"TblGetRowContainingPoint row=%d (x=%d, y=%d)\n", row, x, y));
	return(row);
}

/****************************************************************************
 * Name : GetObjectPtr
 * Desc : get a pointer to an object of the current form
 * Parm : 
 * 			-> ID of object to get a pointer to
 * Out  : ptr
 * Auth : lb, 06.08.2000
 ***************************************************************************/
MemPtr GetObjectPtr(UInt16 ObjID)
{
	FormPtr frm = FrmGetActiveForm();
	return FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ObjID));
} // MemPtr GetObjectPtr(UInt16 ObjID)



/****************************************************************************
 * Name : Today
 * Desc : return today's date
 * Parm : 
 * Out  : today's date
 * Auth : lb, 06.08.2000
 ***************************************************************************/
DateType Today(void)
{
	DateType date;

	DateSecondsToDate(TimGetSeconds(), &date);
	return date;
} // DateType Today(void)



/****************************************************************************
 * Name : DateInLimit
 * Desc : return true if passed date is within limit
 * Parm : 
 * 			-> date to test
 * 			-> offset in limits tab
 * Out  : date is in limit
 * Auth : lb, 04.10.2000, from a modified version of seagull's original code
 ***************************************************************************/
Boolean DateInLimit(DateType date, UInt8 offset)
{
	static UInt8 limits[] = { 0, 0, 1, 2, 3, 4, 5, 6, 7, 14, 21, 28 }; // UInt8 is to saving memory.
	UInt32 days = DateToDays(date);
	Boolean overdue = (days < gTodayDays);
	UInt32 diff = days > gTodayDays ? 
		days - gTodayDays : gTodayDays - days;

	if (!offset)
		return false;

	return ((diff < (UInt32)limits[offset]) || overdue);
} // Boolean DateInLimit(DateType date, UInt8 offset)



/****************************************************************************
 * Name : PrintDueDate
 * Desc : print the date at given y pos
 * Parm : 
 * 			-> the record
 * 			-> the y position
 * 			<-> the x position (in = right pos, out = left pos)
 * Out  : 
 * Auth : lb, 29.08.2000
 * Mod	: lb, 2001-09-06
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
void PrintDueDate(TaskStandardFields *p, UInt16 *datePos, UInt16 y)
{
	Char dateStrMem[dowLongDateStrLength+1]; // complete date + '!' + '\0'
	Char* dateStr = dateStrMem;
	UInt16 dateLen;

	DateToAscii(p->dueDate.month, p->dueDate.day, 
		p->dueDate.year + YEAR_OFFSET, gDateFormat, dateStrMem); // start 1904

	// calc the date length
	if (!gProjectPrefs.displayYear)
	{
		switch(gDateFormat)
		{
			case dfYMDWithSlashes:
			case dfYMDWithDots:
			case dfYMDWithDashes:
// +++ CHECK THIS +++
// This does not look right! Isn't there a better way to do this?
// +++ CHECK THIS +++
				dateStr += 3;
				break;
			default:
				dateStr[StrLen(dateStr) - 3] = '\0';
				break;
		}
	}

	dateLen = StrLen(dateStr);

	if (gTodayDays > DateToDays(p->dueDate))
	{
		// if the task is completed, don't write the !
		if (p->completed != 10)
		{
			dateStr[dateLen++] = '!';
			dateStr[dateLen] = '\0';
		}
	}

	*datePos -= FntCharsWidth(dateStr, dateLen);

	WinDrawChars(dateStr, dateLen, *datePos, y); // just draw days and month
} // void PrintDueDate(TaskStandardFields *p, UInt16 &datePos, UInt16 y)



/****************************************************************************
 * Name : SetFormTitle
 * Desc : set the title of the given form
 * Parm : 
 * Out  : 
 * Auth : lb, 31.08.2000
 ***************************************************************************/
void SetFormTitle (FormPtr frm, Char *title)
{
	#define titleArraySize 33
	Char titleArray[titleArraySize];
	UInt8 i;

	// exit if sanity check fails
	if (NULL == title)
	{
		DBGMSG((DBB, "SetFormTitle - Hey!, What do you think you are "
			"doing! You just passed a NULL pointer for title"));
		return;
	}

	// if prepended, skip prepend
	if (StrStr(title, gPrepend) == title)
	{
		title += 5;
	}
	for (i = 0; i < (titleArraySize -1) && *title; i++)
	{
		titleArray[i] = *title++;
	}
	titleArray[i] = '\0';
	FrmCopyTitle(frm, titleArray);
} // void SetFormTitle (FormPtr frm, Char *title)



/****************************************************************************
 * Name : DrawWrappedText
 * Desc : Draw text block, with word wrap in the current font.
 *        The first line can optionally be indented on both left & right.
 * Parm :
 * 			-> string to draw
 * 			-> x and y top left corner of block
 *          -> lineLimit: Max # of lines to word wrap.
 * 			-> widthLimit: (pixels) width of the block to draw in.
 *          -> indentLeft, indentRight: Amount to indent first line only
 *          -> strikeThrough: If true, a line will be drawn through text.
 * Out  :   <- The number of lines of text actually drawn.
 * Auth : hcc, 05.05.2001, rprice 07.25.2003
 ***************************************************************************/
int DrawWrappedText(Char *strP, UInt16 x, UInt16 y,
					UInt16 lineLimit,
					UInt16 widthLimit,
					UInt16 indentLeft,
					UInt16 indentRight,
					Boolean strikeThrough)
{
	Char *msg = strP;      // Cur position in string to print.
	UInt8 lineCnt;             // Which line we are on. First = 1.
	UInt16 lineLen;            // Number of chars to draw on cur line.
	UInt16 curWidthLimit;      // pixels. width limit for cur line.
	UInt16 curX;               // pixels. Start X position of text in cur line.
	RectangleType rectBounds;  // For erase line's rectangle
	Int16 widthToOffset;

	widthToOffset = 0;
	lineCnt=0;
	while (msg && *msg && (lineCnt < lineLimit)) 
	{
		if(0 == lineCnt)
		{
			curWidthLimit = widthLimit - indentLeft - indentRight;
			curX = x + indentLeft;
		}
		else
		{
			curWidthLimit = widthLimit;
			curX = x;
			RctSetRectangle(&rectBounds, 0, y, OSCaps.screenWidth, 11);
			WinEraseRectangle(&rectBounds, 0);
		}

		widthToOffset = FntWordWrap(msg, curWidthLimit);

		if (lineCnt < lineLimit - 1)
			WinDrawChars(msg, widthToOffset, curX, y);
		else
			// On last line. Trunc middle of word and draw ellipses.
			DrawTruncText(msg, curX, y, curWidthLimit);
		
		if(strikeThrough)
		{
		    lineLen = StrLen(msg);

			// We redefine curWidthLimit to be the width actually used:
			if(lineCnt < lineLimit - 1) // Lines other than last line
				curWidthLimit = FntCharsWidth(msg, lineLen);
			else // Last line, drawn above via DrawTruncText
				FntCharsInWidth(msg, &curWidthLimit, &lineLen,(void *)(&rectBounds)); // last arg is junk ptr

			WinDrawLine(x, y + 6, x + curWidthLimit + (curX - x), y + 6);
		}

		y += FntLineHeight();
		msg += widthToOffset;
		lineCnt++;
	}		

	return (lineCnt);
}

/****************************************************************************
 * Name : DrawTruncText
 * Desc : draw a text, trunc if too long
 * Parm :
 * 			-> string to draw
 * 			-> x and y pos to draw
 * 			-> x limit
 * Out  :  -
 * Auth : lb, 12.09.2000
 * Mod  : hcc, 05.05.2001
 *          Fix ver 3.0 case to consider width of all three dots of ellipsis
 ***************************************************************************/
void DrawTruncText(Char *c, UInt16 x, UInt16 y, UInt16 limit)
{
	Char * ptr = StrChr (c, linefeedChr);
	UInt16 titleLen = (ptr == NULL ? StrLen (c) : (UInt16) (ptr - c));

	if (OSCaps.ver32)
		// best choise. complete nationalized. and got wide screen.
		WinDrawTruncChars(c, titleLen, x, y, limit);
	else if (OSCaps.ver30)
	{   // 99% nationalized. and lost screen width.
		Int16 length = titleLen;
		Int16 len = length;
		Int16 width = limit;
		Boolean fitInWidth = false;
		FntCharsInWidth(c, &width, &len, &fitInWidth);
		if (fitInWidth)
			WinDrawChars(c, length, x, y);
		else
		{
			width = limit - FntCharsWidth("...", 3);
			len = length;
			fitInWidth = false;
			FntCharsInWidth(c, &width, &len, &fitInWidth);
			
			WinDrawChars(c, len, x, y);
			WinDrawChars("...", 3, x + width, y);
		}
	}
	else
	{   // 0% nationalized. but best fit with many PalmOS versions.
		Char *string = c;
		UInt16 textWidth = 0;
		Boolean trunc = false;
		
		while (*c && *c != linefeedChr)
		{
			textWidth += FntCharWidth(*c);
			if (textWidth > limit)
			{
				trunc = true;
				break;
			}
			c++;
		}
		// now textwidth is in chars !!!
		textWidth = (UInt16)(c - string);
		WinDrawChars(string, textWidth, x, y);
		if (trunc)
			WinDrawChars("...", 3, 
						 x + FntCharsWidth(string, textWidth), y);
	}
} // void DrawTruncText(Char *c, UInt16 x, UInt16 y, UInt16 limit)



/****************************************************************************
 * Name : ProjectDrawExtendedItemDesc
 * Desc : draw an item desription
 * Parm :
 * 			-> pointer to the item
 * 			-> x and y pos to draw
 * Out  : -
 * Auth : lb, 12.09.2000
 ***************************************************************************/
void ProjectDrawExtendedItemDesc(TaskExtendedRecordType *p, UInt16 x, UInt16 y)
{
	UInt16 datePos = OSCaps.screenWidth;
	UInt8 textPos = 0;
	UInt16 limit;
	Char* pDesc = NULL;

	// draw note icon (if a note exist)
	if (p->format.bits.hasNote)
	{
		datePos -= 6;
		FntSetFont(symbolFont);
		WinGlueDrawChar(symbolNote, datePos, y);
		FntSetFont(stdFont);
		datePos--;
	}

	// draw link-via icon
	if (p->format.bits.hasLink)
	{
		MemHandle h = gIconHandle[BtmLinkMaster - BtmOffset];
		BitmapPtr bitmap = (BitmapPtr)MemHandleLock(h);
		datePos -= 8;
		WinDrawBitmap(bitmap, datePos, y + 2);
		MemHandleUnlock(h);
		datePos--;
	}

	x += textPos;
	limit = datePos - x - 7;

	//
	// draw attached icon
	//
	{
		UInt16 iconid;
		UInt16 size = sizeof(iconid);
		if (pgOK == TaskGetExtraChunkByTaskPtr(p, Extra_Icon, 0,
											   &iconid, &size) &&
			size == sizeof(iconid) &&
			iconid < getNumIconsLoaded() )
		{
			drawIcon(iconid, x, y);
			x += getIconWidth(iconid) + 2;
		}
	}

	// draw the description (truncated)
	pDesc = TaskGetDescriptionByTaskPtr(p);
	FntSetFont(stdFont);

	// if it's a direct link, and the destination is not existing anymore,
	// nothing to print.
	if (pDesc)
	{
		DrawTruncText(pDesc, x, y, limit);
		MemPtrFree(pDesc);
	}
} // void ProjectDrawExtendedItemDesc(TaskRecordType *p, UInt16 x, UInt16 y)



/****************************************************************************
 * Name : ProjectDrawItemDesc
 * Desc : draw an item desription
 * Parm :
 * 			-> pointer to the item
 * 			-> x and y pos to draw
 * Out  :   <- Number of lines drawn.
 * Auth : lb, 12.09.2000
 * Mod  : hcc, 04.28.2001
 * 		    Support word wrapping. Return number of lines wrapped.
 * 		  lb, 09.08.2001
 * 		  	Fixed icon position bug
 ***************************************************************************/
UInt8 ProjectDrawItemDesc(TaskExtendedRecordType *p, UInt16 x, UInt16 y, 
	UInt8 lineLimit)
{
	UInt16 datePos = OSCaps.screenWidth;
	UInt8 textPos = 0;
	UInt8 leftIndentText = 0;
	UInt16 widthLimit;
	RectangleType rectBounds;
	UInt8 lines = 1;
	TaskStandardFields *pf = StdFields(p);
	Char *desc = NULL;

	// draw note icon (if a note exist)
	if (p->format.bits.hasNote)
	{
		datePos -= 6;
		FntSetFont(symbolFont);
		WinGlueDrawChar(symbolNote, datePos, y);
		datePos--;
	}

	// draw link-via icon
	if (p->format.bits.hasLink)
	{
		MemHandle h = gIconHandle[BtmLinkMaster - BtmOffset];
		BitmapPtr bitmap = (BitmapPtr)MemHandleLock(h);
		datePos -= 8;
		WinDrawBitmap(bitmap, datePos, y + 2);
		MemHandleUnlock(h);
		datePos--;
	}

	// draw the due date
	// if no date, don't try to print a date
	// if DisplayDueDates if false, don't print the date (preference)
	if (pf->dueDate.month != 0 && gProjectPrefs.displayDueDates)
	{
		FntSetFont(stdFont);
		PrintDueDate(pf, &datePos, y);
	}

	//
	// draw progresses
	//
	if (!((gProjectPrefs.hideDoneProgress && 
		 (pf->completed == 10)) ||
		 gProjectPrefs.hideProgress))
	{
		// draw progress item
		if (p->format.bits.itemType == actionType)
		{
			if (OSCaps.ver32 || pf->completed == ACTION_NOT_DONE)
			{
				FntSetFont(symbol11Font);
				WinGlueDrawChar( (pf->completed == ACTION_DONE) ? 
					symbolCheckboxOn : symbolCheckboxOff, x - 3, y);
				FntSetFont(stdFont);
			}
			else
			{
				// get ToDo link icon
				MemHandle h = gIconHandle[BtmCheckedBox - BtmOffset];
				BitmapPtr bitmap = (BitmapPtr)MemHandleLock(h);
				
				// draw it
				WinDrawBitmap(bitmap, x - 2, y);
				MemHandleUnlock(h);
			}
		}
		else if (p->format.bits.itemType == progressType)
		{
			// draw progress bar
			RctSetRectangle(&rectBounds, x - 1, y + 4, 10, 3);
			WinDrawRectangleFrame(simpleFrame, &rectBounds);
			RctSetRectangle(&rectBounds, x - 1, y + 4, pf->completed, 3);
			WinDrawRectangle(&rectBounds, 0);
		}
		else if (p->format.bits.itemType == numericType)
		{
			// draw progress bar
			RctSetRectangle(&rectBounds, x - 1, y + 4, 10, 3);
			WinDrawRectangleFrame(simpleFrame, &rectBounds);
			RctSetRectangle(&rectBounds, x - 1, y + 4, pf->completed, 3);
			WinDrawRectangle(&rectBounds, 0);
		}

		if (p->format.bits.itemType != informativeType)
			textPos += 13;

		if (gProjectPrefs.displayPriorities && pf->priority > 0 && 
			pf->priority < NO_PRIORITY)
		{
			FntSetFont(stdFont);
			WinGlueDrawChar('[', x + textPos, y);
			FntSetFont(boldFont);
			WinGlueDrawChar(pf->priority + '0', x + textPos + 3, y);
			FntSetFont(stdFont);
			WinGlueDrawChar(']', x + textPos + 10, y);
			leftIndentText += 13;
		}
	}

	// High priority task draw bold font.
	//  I wish color to red.
	FntSetFont(stdFont); // Initialy
	if (pf->completed != 10) // don't hilight done task
	{
		if (pf->priority > 0 && pf->priority <= gProjectPrefs.boldMinPriority)
			FntSetFont(boldFont);
		else if (p->format.bits.hasDueDate && gProjectPrefs.boldMinDays != 0 &&
			pf->dueDate.month != 0)
		{
			if (DateInLimit(pf->dueDate, gProjectPrefs.boldMinDays))
			{
//				static RGBColorType red = { 0, 0xff, 0, 0 };
//				WinSetTextColor(WinRGBToIndex(&red));
				FntSetFont(boldFont);
			}
		}
	}

	x += textPos;

	//
	// draw attached icon
	//
	{
		UInt16 iconid;
		UInt16 size = sizeof(iconid);
		if (pgOK == TaskGetExtraChunkByTaskPtr((TaskExtendedRecordType*)p,
											   Extra_Icon, 0,
											   &iconid, &size) &&
			size == sizeof(iconid) &&
			iconid < getNumIconsLoaded() )
		{
			drawIcon(iconid, x + leftIndentText, y + 6 - getIconHeight(iconid) / 2);
			leftIndentText += getIconWidth(iconid) + 2;
		}
	}

	// draw the description (truncated)
	widthLimit = OSCaps.screenWidth - x; //datePos - x; // - 7;

	// HCC: Word-wrapping support added here
	// DrawTruncText(&p->description, x, y, limit);
//	lines = DrawWrappedText(&pf->description, x, y, lineLimit, 
//		widthLimit, leftIndentText, OSCaps.screenWidth - datePos,
//		gProjectPrefs.strikeDoneTasks &&
//		(pf->completed == 10));
	desc = TaskGetDescriptionByTaskPtr(p);
	lines = DrawWrappedText(desc , x, y, lineLimit, 
		widthLimit, leftIndentText, OSCaps.screenWidth - datePos,
		gProjectPrefs.strikeDoneTasks &&
		(pf->completed == 10));
	MemPtrFree(desc);


// Strikethrough is Now handled by DrawWrappedText
#if 0
	// strike out done tasks
	//  but, I needs gray color or gray pattern ...
	if (gProjectPrefs.strikeDoneTasks &&
		(p->completed == 10 || p->completed == ACTION_OK) )
	{
		WinDrawLine(x, y + 6, x + widthLimit, y + 6);
	}
#endif
	
	// resume to default. (to safety)
	FntSetFont(stdFont);

	return (lines);
} // UInt8 ProjectDrawItemDesc(TaskExtendedRecordType *p, UInt16 x, UInt16 y, 
  // UInt8 lineLimit)

/****************************************************************************
 * Name : DrawItemTreeLines
 * Desc : Draw the tree line icons for one line of this node.
 * Parm :
 *          -> pointer to task
 *          -> row: The table row being drawn.
 *          -> x: start x pos for drawing (currently unused; must be zero)
 *          -> y: The y position to draw this row of lines
 *          -> lineNum: For multi-line tasks, which line to draw, start at 0.
 * Out  :   <- The amount of horizontal space used (pixels).
 * Auth : hcc, 04.28.2001
 *          (extracted from lb's code in ProjectItemDraw)
 * Mod  : hcc, 05.05.2001
 *          Do wrapped lines. Handle line drawing disabled mode.
 ***************************************************************************/
UInt16 DrawItemTreeLines(TaskRecordType *p, UInt16 row,
						 UInt16 x, UInt16 y, UInt8 lineNum)
{
	UInt8 level;
	UInt8 i;
	BitmapPtr bitmap;
	UInt16 bitmapId = 0;
	MemHandle h;
	UInt8 shiftX = 0; // Optionally shift icon horizontally (pixels)

	// get its level
	level = gCurrentPrefs.view == flatView ? 1 : (p->attr.bits.level);

	// one bitmap per level of the task
	for (i = gRefLevel + 1; i <= level; i++)
	{
		shiftX = 0;
		if(gProjectPrefs.drawTreeLines)
		{
			// choose the good icon (final icon)
			if (i == level)
			{
				if(lineNum == 0)
				{
					if (p->attr.bits.hasChild)
						bitmapId =
							(p->attr.bits.opened)? BtmMinusAlone: BtmPlusAlone;
					
					else // has no children.
					{
						if (p->attr.bits.hasPrev)
							bitmapId =
								(p->attr.bits.hasNext) ?
								BtmItemMiddle : BtmItemBottom;
						else 
							bitmapId = 
								(p->attr.bits.hasNext) ? 
								BtmItemFirst : BtmItemAlone;
					}
				}
				else // wrapped line
				{
					if (p->attr.bits.hasChild && p->attr.bits.opened)
						bitmapId =
							(p->attr.bits.hasNext) ?
							BtmItemChildNextWrapped : BtmItemChildWrapped;
					else
						bitmapId =
							(p->attr.bits.hasNext) ? BtmItemNo : BtmNull;
				}
			}
			else
			{
				if (!(p->attr.bits.hasPrev) && i == level - 1 && lineNum == 0)
					bitmapId = (wVerticalLines[row] & (1 << (i - 1))) ? 
						BtmItemChild : BtmItemChildNo;
				// if there is a bitmap to draw
				// init bitmapId to BtmNull if prev of level i has no next
				else 
					bitmapId = (wVerticalLines[row] & (1 << (i - 1))) ?
						BtmItemNo: BtmNull;
			}
		}
		else // Tree lines are off
		{
			if (i == level && lineNum == 0)
			{
				if (p->attr.bits.hasChild) // Draw + or - icon
					bitmapId =
						(p->attr.bits.opened)? BtmMinusAlone: BtmPlusAlone;
				else // Just draw a bullet.
					bitmapId = BtmItemAloneNoLine;
			}
		}

		if (bitmapId != BtmNull)
		{
			// ok, draw this icon
			// retreive the good bitmap
			h = gIconHandle[bitmapId - BtmOffset];
			bitmap = (BitmapPtr)MemHandleLock(h);

			// draw it
			WinDrawBitmap(bitmap, (i - 1 - gRefLevel) * 11, y);
			MemHandleUnlock(h);

			// indicate ToDo Linking
			// TODO : test code under 3.5
			if (i == level && p->format.bits.hasToDo)
			{
				if (! OSCaps.ver35)
				{
					WinDrawLine((i - 1 - gRefLevel) * 11 + 8, y+5, 
						(i - 1 - gRefLevel) * 11 + 8, y+5);
				}
				else
				{
					WinDrawPixel((i - 1 - gRefLevel) * 11 + 8, y+5);
				}
			}

		}
	}

	// time optimizations
	// TODO : recalc these
	x = (level - gRefLevel) * 11 + 2;

	return(x);
}

/****************************************************************************
 * Name : ProjectItemDraw
 * Desc : callback to draw the project table
 * In   : not called by the user, it's a callback
 * Out  : -
 * Auth : lb, 31.07.2000
 * Mod  : lb, 06.08.2000
 * 		  	support for date right aligning
 * 		  lb, 29.01.2001
 * 		  	fill node when ToDo Link is active
 *        hcc, 05.05.2001
 *          Move line & icon code to new DrawItemTreeLines().
 *          Implement word wrapping option.
 ***************************************************************************/
void ProjectItemDraw(MemPtr table, UInt16 row, UInt16 column, RectanglePtr bounds)
{
	BitmapPtr bitmap;
	MemHandle h, taskH;
	UInt16 id, x = 0;
	UInt16 y = bounds->topLeft.y; //HCC: was: row*11 + 15;
	TaskExtendedRecordType *p;
	FormType *frm;
	RectangleType rectBounds;
	UInt8 lines = 1;

	// If this row exceeds the vertical bounds of the table, bail
	if(y > gProjectTableMaxY - 11)
	{
		TblSetRowHeight(table, row, 0);
		return;
	}

	id = TblGetItemInt(table, row, column);

	frm = FrmGetActiveForm();

	// erase the row
	RctSetRectangle(&rectBounds, 0, y, OSCaps.screenWidth, 11);
	WinEraseRectangle(&rectBounds, 0);

	// don't do anything more if id is -1
	if (id == -1)
	{
		// Row height may have been left higher from previous update. Set it
		// back down so that all remaining rows will be erased.
		TblSetRowHeight(table, row, 11);
		return;
	}

	// get the task
	taskH = DmQueryRecord(gdbP, id);
	p = MemHandleLock(taskH);

#if 0
	// HCC: Debug
	DmRecordInfo(gdbP, id, NULL, &uniqueID, NULL);
	DBGMSG((DBB, "Row:%d Lvl:%d Idx:%d Uid:%ld, bounds y=%d, h=%d\n",
			row, TaskGetLevel(gdbP, id), id, uniqueID,
			bounds->topLeft.y, bounds->extent.y));
#endif

	if (gCurrentPrefs.view == flatView)
	{
		// extended task can't link to ToDo. because it's buggy.
		if (p->format.bits.itemType != extendedType)
		{
			// get ToDo link icon
			h = gIconHandle[
				(p->format.bits.hasToDo? BtmLinkOn: BtmLinkOff) - BtmOffset];
			bitmap = (BitmapPtr)MemHandleLock(h);
			
			// draw it
			WinDrawBitmap(bitmap, 1, y + 2);
			MemHandleUnlock(h);
		}
		x = 13;
	}
#if 0
	// test implement... but this is BAD idea...
	else if (gProjectPrefs.nIndentWidth <= 1)
	{   // for short or very short indent
		x = (level - gRefLevel - 1) * 2;

		if (p->attr.bits.hasChild || gProjectPrefs.nIndentWidth == 1)
		{
			h = gIconHandle[(p->attr.bits.opened? BtmMinusAlone : BtmPlusAlone) - BtmOffset];
			bitmap = (BitmapPtr)MemHandleLock(h);
			if (p->attr.bits.hasChild)
				WinDrawBitmap(bitmap, x, y);
			x += bitmap->width + 1;
			MemHandleUnlock(h);
		}
	}
#endif
	else
	{
		// we can do this just because we only use attr and format
		x = DrawItemTreeLines((TaskRecordType*)p, row, x, y, 0);
	}

	// write the item description (with priority, due date, progress bar...)
	if (p->format.bits.itemType == extendedType)
	{
		ProjectDrawExtendedItemDesc((TaskExtendedRecordType*)p, x, y);
	}
	else
	{
		// Init lines to be limit of the # of lines to show for this item
		lines = (gProjectTableMaxY - y) / 11;
		if(lines > gProjectPrefs.wordWrapLines)
			lines = gProjectPrefs.wordWrapLines;

		// Now lines will be the actual # of lines that were used.
		lines = ProjectDrawItemDesc(p, x, y, lines);
	}

	if (lines * 11 != bounds->extent.y)
	{
		TblSetRowHeight(table, row, lines * 11);
	}

	// Draw continuation vertical lines for each additional
	// horizontal line belonging to this item.
	while (--lines > 0)
	{
		DrawItemTreeLines((TaskRecordType*)p, row, 0, y + 11*lines, lines);
	}

	MemHandleUnlock(taskH);

} // static void ProjectItemDraw(MemPtr table, UInt16 row, UInt16 column, RectanglePtr bounds)



/*
 * Table init
 */
void ProjectTableInit(void)
{
	TablePtr table;
	UInt16 numRows;
	UInt16 i;
	RectangleType tableBounds;
	DBGMSG((DBB, "In ProjectTableInit"));

	table = GetObjectPtr(ProjectTable);
	numRows = TblGetNumberOfRows(table);

	// Cache table's max Y for frequent use by ProjectItemDraw
	TblGetBounds(table, &tableBounds);
	gProjectTableMaxY = tableBounds.topLeft.y + tableBounds.extent.y;

	for (i = 0; i < numRows; i++)
	{
		TblSetItemStyle(table, i, 0, customTableItem);
		TblSetItemInt(table, i, 0, i + 1); // begin at task 1, never display task 0
		TblSetRowSelectable(table, i, false);
		TblSetRowUsable(table, i, false);
	}

	TblSetColumnUsable(table, 0, true);
	TblSetCustomDrawProcedure(table, 0, (TableDrawItemFuncPtr)ProjectItemDraw);
	TblSetItemInt(table, 0, 0, gCurrentPrefs.topTask);
} // static void ProjectTableInit(void)



void ProjectTableEmpty(void)
{
	TablePtr table;
	UInt16 numRows;
	UInt16 i;

	table = GetObjectPtr(ProjectTable);
	numRows = TblGetNumberOfRows(table);

	for (i = 0; i < numRows; i++)
	{
		// fill with empty items
		TblSetItemInt(table, i, 0, -1);
		TblSetRowSelectable(table, i, false);
		TblSetRowUsable(table, i, false);
		wVerticalLines[i] = 0;
	}
}



void ProjectTableUpdate(void)
{
	TablePtr table;
	UInt16 numRows;
	UInt16 numRec;
	UInt16 i;
	Int32 next;
	UInt8 level;
	DBGMSG((DBB, "In ProjectTableUpdate"));

	table = GetObjectPtr(ProjectTable);
	numRows = TblGetNumberOfRows(table);
	numRec = PgNumRecords(gdbP);

	// need to set the first one selectable for when we come back of an empty
	// subview
	TblSetRowSelectable(table, 0, true);
	TblSetRowUsable(table, 0, true);

	// special case, db empty (just task 0)
	if (numRec == 1)
	{
		ProjectTableEmpty();
		return;
	}

	// special case, the first row's item is -1, check
	if (TblGetItemInt(table, 0, 0) == -1 || 
		TblGetItemInt(table, 0, 0) >= numRec)
	{
		if (gParentTask + 1 < numRec && 	
			TaskGetLevel(gdbP, gParentTask + 1) > gRefLevel)
		{
			TblSetItemInt(table, 0, 0, gParentTask + 1);
			TblSetRowSelectable(table, 0, true);
			TblSetRowUsable(table, 0, true);
		}
		else
		{
			ProjectTableEmpty();
			return;
		}
	}

	// first item change if it is now invisible
	while (gProjectPrefs.hideDoneTasks && (next = TblGetItemInt(table, 0, 0)) != -1 && TaskGetIsDone(gdbP, next))
	{
		next = TaskGetNextRelativeIndex(gdbP, next);
		if (next == 0)
		{
			ProjectTableEmpty();
			return;
		}
		TblSetItemInt(table, 0, 0, next);
		TblSetRowSelectable(table, 0, true);
		TblSetRowUsable(table, 0, true);
	}

	if (TaskGetLevel(gdbP, TblGetItemInt(table, 0, 0)) <= gRefLevel)
	{
		ProjectTableEmpty();
		return;
	}

	
	//
	// calc vertical line for first row.
	//
	// HCC NOTE: This loop for the first node is where most of the
    // performance penalty lies, because TaskGetPrevIndexByLevel can be
	// an expensive call.
	if(gProjectPrefs.drawTreeLines)
	{
		next = TblGetItemInt(table, 0, 0);
		level = TaskGetLevel(gdbP, next);
		wVerticalLines[0] = TaskGetHasNext(gdbP, next)? (1 << (level - 1)) : 0;
		for (i = gRefLevel + 1; i < level; i++)
		{
			UInt16 prev;
			if ( ( !TaskGetHasPrev(gdbP, next) && i == level - 1 && TaskGetHasNext(gdbP, next - 1) ) || 
				 ( (prev = TaskGetPrevIndexByLevel(gdbP, next, i)) != dmMaxRecordIndex && TaskGetHasNext(gdbP, prev) ) )
				wVerticalLines[0] |= 1 << (i - 1);
		}
	}

	// other items
	for (i = 1; i < numRows; i++)
	{
		// get the previous item
		next = TblGetItemInt(table, i-1, 0);

		if (next != -1)
		{
			// if he is not opened
			if (TaskGetHasChild(gdbP, next) && !TaskIsOpened(gdbP, next))
			{
				// get the next
				next = TaskGetNextRelativeIndex(gdbP, next);
				if (next == 0)
					next = -1;
			}
			else
			{
				// just go to the next
				next++;

				// if it's past the last record
				if (next >= numRec)
					next = -1;
			}

			// if he must not be drawn
			while (gProjectPrefs.hideDoneTasks && next > 0 && TaskGetIsDone(gdbP, next))
				next = TaskGetNextRelativeIndex(gdbP, next);
			if (next <= 0)
				next = -1;

			// if we are drawing a subtree, stop when we reach the last child
			if (gParentTask != 0 && next != -1)
			{	
				UInt8 levA;
				if ((levA = TaskGetLevel(gdbP, next)) <= (gRefLevel) ||
					(  levA == gRefLevel + 1 &&
					   TaskGetFatherIndex(gdbP, next) != gParentTask)  )
					next = -1;
			}

		}

		// calc vertical line for each row.
		if(gProjectPrefs.drawTreeLines)
		{
			if (next == -1)
				wVerticalLines[i] = 0;
			else 
			{
				UInt16 mask = 1 << (TaskGetLevel(gdbP, next) - 1);
				if (TaskGetHasNext(gdbP, next))
					wVerticalLines[i] = wVerticalLines[i - 1] | mask;
				else
					wVerticalLines[i] = wVerticalLines[i - 1] & ~mask;
			}
		}

		TblSetItemInt(table, i, 0, next);
		TblSetRowSelectable(table, i, (next >= 0));
		TblSetRowUsable(table, i, (next >= 0));
	}
	gCurrentPrefs.topTask = TblGetItemInt(table, 0, 0);
} // void ProjectTableUpdate(void)



/****************************************************************************
 * Name : Scroll
 * Desc : update the form to scroll up or down
 * Parm : 
 * 			-> scroll type (scrollUp, scrollDown)
 * 			-> scroll step
 * Out  : 
 * Auth : lb, 02.08.2000
 * Mod  : lb, 18.08.2000
 * 			fixed scroll up bug that displayed father in sub view
 ***************************************************************************/
void Scroll(scrollType s, UInt8 step)
{
	UInt16 value;
	TablePtr table;

	table = GetObjectPtr(ProjectTable);

	switch (s)
	{
		case scrollDown:
			if ((value = TblGetItemInt(table, step, 0)) != -1)
			{
				TblSetItemInt(table, 0, 0, value);
				//FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
				ProjectTableUpdate();
			}
			else
			{
				step--;
				while (step > 0)
				{
					if ((value = TblGetItemInt(table, step, 0)) != -1)
					{
						TblSetItemInt(table, 0, 0, value);
						//FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
						ProjectTableUpdate();
						break;
					}
					step--;
				}
			}
			break;

		case scrollUp:
			value = TblGetItemInt(table, 0, 0);
			if (value == -1)
				return;
			if (value > step + gParentTask)
			{
				UInt8 count = step;
				while (value > gParentTask + 1 && count)
				{
					value--;
					if (TaskIsVisible(gdbP, value))
					{
						count--;
					}
					else
					{
						if (gProjectPrefs.hideDoneTasks)
							value--;
						else
							value = TaskGetFatherIndex(gdbP, value);
						if (value == gParentTask)
						{
							value = gParentTask + 1;
						}
						if (TaskIsVisible(gdbP, value))
						{
							count--;
						}
					}
				}
				TblSetItemInt(table, 0, 0, value);
				//FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
				ProjectTableUpdate();
			}
			else
			{
				TblSetItemInt(table, 0, 0, gParentTask + 1); // never display task 0
				ProjectTableUpdate();
				//FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			}
			break;

		case scrollTop:
			if (TaskGetHasChild(gdbP, gParentTask))
				TblSetItemInt(table, 0, 0, gParentTask + 1);
			else
				TblSetItemInt(table, 0, 0, -1);
			// update done in Btn_Sub
			//FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			break;

		default:
			break;
	}
} // void Scroll(scrollType s, UInt8 step)



/****************************************************************************
 * Name : FrmTaskEditUpdateDateTrigger
 * Desc : update the date trigger
 * Parm : 
 * 			-> trigger pointer
 * 			-> date to print
 * Out  : 
 * Auth : lb, 06.08.2000
 ***************************************************************************/
void FrmTaskEditUpdateDateTrigger(MemPtr trigger, DateType p)
{
	// +++ FIX THIS +++ 
	// This code should probably allocate a new buffer, or actually
	// maybe we can just use a static buffer here. We need to manage
	// the memory for it in any case.
	// This code does not seem reliable.
	// +++ FIX THIS +++ 

	Char* date;
	date = (Char*)CtlGetLabel(trigger);

	if (p.month == 0)
	{
		StrCopy(date, "No date");
	}
	else
	{
		DateToAscii(p.month, p.day, 
			p.year + YEAR_OFFSET, PrefGetPreference(prefDateFormat), date);
	}



	CtlSetLabel(trigger, date);
} // void FrmTaskEditUpdateDateTrigger(MemPtr trigger, DateType p)



/****************************************************************************
 * Name : FrmTaskEditUpdateDateTrigger
 * Desc : update the date trigger
 * Parm : 
 * 			-> trigger pointer
 * 			-> date to print
 * Out  : 
 * Auth : lb, 06.08.2000
 ***************************************************************************/
void UpdatePopupTrigger(MemPtr trigger, MemPtr list)
{
	Char* text;

	// +++ FIX THIS +++ 
	// This code should probably allocate a new buffer, or actually
	// maybe we can just use a static buffer here. We need to manage
	// the memory for it in any case.
	// This code does not seem reliable.
	// +++ FIX THIS +++ 

	text = (Char*)CtlGetLabel(trigger);
	StrCopy(text, LstGetSelectionText(list, LstGetSelection(list)));
	CtlSetLabel(trigger, text);
} // void UpdatePopupTrigger(MemPtr trigger, MemPtr list)



/****************************************************************************
 * Name : SelectActualTask
 * Desc : select the tbl case of gActualTask. Move the screen if necessary
 * 		  expand a branch if necessary
 * Parm : 
 * 			-> move (must the actual task be on screen)
 * Out  : 
 * Auth : lb, 09.08.2000
 * Mod  : hcc, 05.05.2001
 *          Make compatible with word wrapping.
 ***************************************************************************/
void SelectActualTask(Boolean move)
{
	TablePtr table = GetObjectPtr(ProjectTable);
	UInt8 i = 0, numRows = TblGetNumberOfRows(table);

	DBGMSG((DBB, "In SelectActualTask : %hu", gActualTask));
	gLastSelected = 0;
	TblUnhighlightSelection(table);

	if (gActualTask == 0 || gActualTask > PgNumRecords(gdbP))
		return;

	// ensure the branch is expanded
	if (move && !TaskIsVisible(gdbP, gActualTask))
			TaskMakeVisible(gdbP, gActualTask);

	// is actual task on screen ?
	while (i < numRows && TblGetItemInt(table, i, 0) != gActualTask)
	{
		i++;
	}

	// HCC: Test if it is actually on screen, since multi-line items
    // may push some table rows off bottom of table.
	if (i != numRows && TblGetRowHeight(table, i) == 0)
	{
		i = numRows;
	}

	if (i == numRows && move)
	{
		TblSetItemInt(table, 0, 0, gActualTask);
		// center it
		// HCC: Workaround to keep item on screen when using word wrapping.
		// Since I can't determine the height of each row without drawing it,
		// Assume each row is the max number of lines.
		if (gProjectPrefs.wordWrapLines < 1) // Paranoia
			gProjectPrefs.wordWrapLines = 1;
		Scroll(scrollUp, 11 / gProjectPrefs.wordWrapLines);
		// where is it now
		i = 0;
		while (i < numRows && TblGetItemInt(table, i, 0) != gActualTask)
		{
			i++;
		}
	}
	if (i != numRows)
	{
		TblSelectItem(table, i, 0);
		gLastSelected = gActualTask;
	}
	else
	{
		gLastSelected = gActualTask = 0;
	}
} // void SelectActualTask(Boolean move)



/****************************************************************************
 * Name : InitFrmTaskEdit
 * Desc : init the form to edit the actual selected task
 * Parm : 
 * Out  : 
 * Auth : lb, 03.08.2000
 * Mod  : rw, 30.12.2000
 * 		  	support for syncing categories with ToDo
 * Mod  : rw, 02.01.2001
 * 		  	modified to use ItemType enumeration values
 * Mod  : lb, 23.06.2001
 * 			handera support : correct date trigger position
 * Mod	: lb, 2001-09-06
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
void InitFrmTaskEdit(void)
{
	FieldPtr fld = GetObjectPtr(TaskDesc);
	MemHandle h;
	TaskExtendedRecordType *p;
	MemHandle textH = FldGetTextHandle(fld);
	Char *textP;
	UInt16 l;
	UInt8 priority, value;
	DateType dueDate;
	UInt16 category;
	ItemType type;
	TaskStandardFields *pf;

	// if there's already a handle, destroy it
	if (textH)
	{
		FldSetTextHandle(fld, NULL);
		MemHandleFree(textH);
	}

	h = DmQueryRecord(gdbP, gActualTask);
	p = MemHandleLock(h);
	pf = StdFields(p);
	l = StrLen(&pf->description);
	textH = MemHandleNew(l+1);
	textP = MemHandleLock(textH);
	SafeCopy(textP, &pf->description,l+1);
	MemHandleUnlock(h);
	MemHandleUnlock(textH);
	FldSetTextHandle(fld, textH);

	// see if we're synchronized with a ToDo
	if (TaskGetFormat(gdbP, gActualTask).hasToDo)
	{
		pgErr err;
		// check the status in ToDo and update this task
		err = TaskGetToDoStatus(gdbP, gActualTask, &value, &priority, &dueDate, &category);
		if (err == pgOK)
		{
			TaskSetCompleted(gdbP, gActualTask, value, true);
			TaskSetDueDate(gdbP, gActualTask, dueDate);
			TaskSetPriority(gdbP, gActualTask, priority);
			TaskSetCategory(gdbP, gActualTask, category);
			CtlSetValue(GetObjectPtr(Chk_Edit_ToDo_Link), 1);
		}
	}

	// get the date and set the selector trigger
	FrmTaskEditUpdateDateTrigger(GetObjectPtr(Pop_Due_Date), TaskGetDueDate(gdbP, gActualTask));

	// get the priority and set it on the form
	priority = TaskGetPriority(gdbP, gActualTask);
	// to be compatible with not initialised db
	if (priority > 0 && priority < 7)
	{
		CtlSetValue(GetObjectPtr(Btn_Priority_X + priority - 1), true);
	}

	// initialize category popup trigger
	{
		Char* name = (Char*)CtlGetLabel(GetObjectPtr(Pop_Category));
		category = TaskGetCategory(gdbP, gActualTask);
		CategoryGetName(gdbP, category, name);
		CategorySetTriggerLabel(GetObjectPtr(Pop_Category), name);
	}

	// initialize item type list and popup trigger
	type = TaskGetType(gdbP, gActualTask);

	// +++ FIX THIS +++
	// Okay, so now what if the type is extendedType or linkType(not supported)?
	// It'll blow up the GUI.
	// This should be fixed somehowe should it not?
	// +++ FIX THIS +++

	LstSetSelection(GetObjectPtr(Lst_Item_Type), type);
	if (type == numericType)
	{
		UInt32 values;
		UInt16 size = sizeof(values);

		Char actualstr[16], limitstr[16];
		UInt16 actual=0, limit=0;

		// show value fields
		FrmTaskEditPart(showValue);

		if (pgOK == TaskGetExtraChunk(gdbP, gActualTask, Extra_Numeric, 0,
						  &values, &size))
		{
			actual = (UInt16)(values & 0x0000FFFF);
			limit  = (UInt16)(values >> 16 & 0x0000FFFF);
		}

		StrIToA(actualstr, actual);
		StrIToA(limitstr, limit);

		FldInsert(GetObjectPtr(TaskNumericActual), actualstr, 
			StrLen(actualstr));
		FldInsert(GetObjectPtr(TaskNumericLimit), limitstr, 
			StrLen(limitstr));
	}

	UpdatePopupTrigger(GetObjectPtr(Pop_Item_Type), 
		GetObjectPtr(Lst_Item_Type));

} // void InitFrmTaskEdit(void)



/****************************************************************************
 * Name : CleanUpFrmTaskEdit
 * Desc : clean up the form, save the work done
 * Parm : 
 * Out  : 
 * Auth : lb, 03.08.2000
 * Mod	: lb, 2001-09-10
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
Boolean CleanUpFrmTaskEdit(void)
{
	FieldPtr fld = GetObjectPtr(TaskDesc);
	MemHandle textH = FldGetTextHandle(fld);
	Char *textP;
	TaskFormatType format;
	Boolean bDeleteTask=false;

	FldCompactText(fld);
	textP = MemHandleLock(textH);

	// if the string length is zero, we need to delete this task when this is all done :(
	if (StrLen(textP) == 0)
		bDeleteTask=true;

	DBGMSG((DBB, "In CleanUpFrmTaskEdit"));
	dbgPrintTask(gdbP, gActualTask);
	TaskSetDescription(gdbP, gActualTask, textP);
	MemHandleUnlock(textH);

	// update or create extrablock for numeric type
	format = TaskGetFormat(gdbP, gActualTask);
	format.newTask = 0;
	TaskSetFormat(gdbP, gActualTask, format);
	if (format.itemType == numericType)
	{
		Char *fieldTextPointer;
		UInt32 values, percentage;
		UInt16 limit=0, actual=0; // default to zero

		// fieldTextPointer _can_ be NULL, so assume default if it is....
		fieldTextPointer = FldGetTextPtr(GetObjectPtr(TaskNumericLimit)); 
		if (fieldTextPointer)
			limit = (UInt16)StrAToI(fieldTextPointer);

		values = limit;

		values <<= 16; // put it on the right

		// add the actual to the left
		// fieldTextPointer _can_ be NULL, so assume default if it is....
		fieldTextPointer = FldGetTextPtr(GetObjectPtr(TaskNumericActual)); 
		if (fieldTextPointer)
			actual =(UInt16)StrAToI(fieldTextPointer);

		values |= actual;

		TaskSetExtraChunk(gdbP, gActualTask, Extra_Numeric, 0,
						  &values, sizeof(UInt32));

		if (limit > 0)
		{
			percentage = (actual * 10) / limit;
			if (percentage > 10)
				percentage = 10;
		}
		else
			percentage = 0;

		TaskSetCompleted(gdbP, gActualTask, (UInt16)percentage, true);

	}
	else // not numeric, delete numeric extrablock
	{
		TaskRemoveExtraChunk(gdbP, gActualTask, Extra_Numeric, 0);
	}

	return !bDeleteTask;

} // Boolean CleanUpFrmTaskEdit(void)



/****************************************************************************
 * Name : InitFrmNoteEdit
 * Desc : init the form to edit the actual selected task
 * Parm : 
 * Out  : 
 * Auth : lb, 22.08.2000
 ***************************************************************************/
void InitFrmNoteEdit(void)
{
	FieldPtr fld = GetObjectPtr(TaskNote);
	MemHandle textH = FldGetTextHandle(fld);
	Char *textP;
	UInt16 l;
	Char *note;

	DBGMSG((DBB, "InitFrmNoteEdit"));

	// if there's already a handle, destroy it
	if (textH)
	{
		FldSetTextHandle(fld, NULL);
		MemHandleFree(textH);
	}

	note = TaskGetNote(gdbP, gActualTask);
	// if there's a description already
	if (note)
	{
		l = StrLen(note);
		textH = MemHandleNew(l+1);
		textP = MemHandleLock(textH);

		// could change as we allocated memory
		note = TaskGetNote(gdbP, gActualTask);
		SafeCopy(textP, note,l+1);
		MemHandleUnlock(textH);
		FldSetTextHandle(fld, textH);
	}
	else
	{
		// set as empty string.
		Char* p;
		textH = MemHandleNew(1);
		p = MemHandleLock(textH);
		*p = '\0';
		MemHandleUnlock(textH);
		FldSetTextHandle(fld, textH);
	}

	DBGMSG((DBB, "InitFrmNoteEdit Exit"));

} // void InitFrmNoteEdit(void)



/****************************************************************************
 * Name : CleanUpFrmNoteEdit
 * Desc : clean up the form, save the work done
 * Parm : 
 * Out  : 
 * Auth : lb, 22.08.2000
 ***************************************************************************/
void CleanUpFrmNoteEdit(void)
{
	FieldPtr fld = GetObjectPtr(TaskNote);
	MemHandle textH = FldGetTextHandle(fld);
	Char *textP;

	FldCompactText(fld);

	textP = MemHandleLock(textH);

	TaskSetNote(gdbP, gActualTask, textP);

	MemHandleUnlock(textH);
	
	// set texthandle as NULL, or it'll be destroyed by the close event
	FldSetTextHandle(fld, NULL);
	MemHandleFree(textH);
} // void CleanUpFrmNoteEdit(void)


/****************************************************************************
 * Name : MessageBox
 * Desc : display a message for the user, just OK
 * Parm :
 * 			-> message id
 * Out  : 
 * Auth : lb, 11.09.2000
 ***************************************************************************/
void MessageBox(UInt16 id)
{
  MemHandle h;
  char* p;
  UInt16 respond;

  h = DmGetResource('tSTR', id);
  p = MemHandleLock(h);

  respond = FrmCustomAlert(AltEmpty, p, " ", " ");

  MemHandleUnlock(h);
  DmReleaseResource(h);

} // void MessageBox(UInt16 id)



/****************************************************************************
 * Name : GetLastInSub
 * Desc : return the last index of the parent's sub
 * Parm :
 * 			-> parent
 * Out  : 
 * Auth : lb, 21.10.2000
 ***************************************************************************/
UInt16 GetLastInSub(UInt16 parent)
{
	UInt16 i = parent + 1;
	UInt8 refLevel = (parent == 1 ? 0 : TaskGetLevel(gdbP, parent));
	UInt16 num = PgNumRecords(gdbP) - 1;
	while (i < num && TaskGetLevel(gdbP, i) > refLevel)
	{
		i++;
	}
	return i - 1;
} // UInt16 GetLastInSub(UInt16 parent)



/****************************************************************************
 * Name : ConfirmCustom
 * Desc : display a custom confirmation message
 * Parm :
 * 			-> action id (StrActionDelete || StrActionResetDates
 * 			-> message id
 * Out  : button pressed
 * Auth : seagull, 09.09.2000
 * Mod  : burgbach, 21.10.2000 -> custom message (not only delete messages)
 ***************************************************************************/
UInt16 ConfirmCustom(UInt16 idAction, UInt16 id)
{
  MemHandle h1, h2;
  char *p1, *p2;
  UInt16 respond;

  h1 = DmGetResource('tSTR', idAction);
  h2 = DmGetResource('tSTR', id);
  p1 = MemHandleLock(h1);
  p2 = MemHandleLock(h2);

  respond = FrmCustomAlert(AltConfirmCustom, p1, p2, " ");

  MemHandleUnlock(h1);
  DmReleaseResource(h1);
  MemHandleUnlock(h2);
  DmReleaseResource(h2);

  return respond;
} // UInt16 ConfirmCustom(UInt16 idAction, UInt16 id)



/****************************************************************************
 * Name : ProgressHandleEvent
 * Desc : update the progress of the actualtask
 * Parm :
 * Out  : 
 * Auth : burgbach, 19.10.2000
 * Mod  : rw, 30.12.2000
 * 		  	updated signature of TaskSetToDoStatus
 * Mod	: lb, 2001-09-06
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
void ProgressHandleEvent(void)
{
	// only for leaves
	// not for directlink types
	TaskFormatType format;

	if (!gActualTask)
		return;

	format = TaskGetFormat(gdbP, gActualTask);
	if (format.itemType == actionType)
	{
		TaskSetCompleted(gdbP, gActualTask, 
			TaskGetCompleted(gdbP, gActualTask) == ACTION_DONE ? 
			ACTION_NOT_DONE : ACTION_DONE, true);
		if (format.hasToDo)
		{
			TaskSetToDoStatus(gdbP, gActualTask, toDoComplete,
				TaskGetCompleted(gdbP, gActualTask), 0,
				gNoDate, 0);
			// record completion date (always for ToDo linked tasks)
			if (TaskGetIsDone(gdbP, gActualTask))
				TaskSetToDoStatus(gdbP, gActualTask, toDoDueDate,
					0, 0,
					Today(), 0);
		}
	}
	else if (!TaskGetHasChild(gdbP, gActualTask) && 
		(format.itemType == progressType))
	{
		Int16 TInt16;
		ListPtr listP = GetObjectPtr(Lst_Percent);
		LstSetSelection(listP, 
			TaskGetCompleted(gdbP, gActualTask));

		TInt16 = LstPopupList(listP);
		if (TInt16 != -1)
		{
			TaskSetCompleted(gdbP, gActualTask, TInt16, true);
		}
	}
	else
		return;

	// record completion date
	if (TaskGetIsDone(gdbP, gActualTask) && gProjectPrefs.completionDate)
		TaskSetDueDate(gdbP, gActualTask, Today());

	// update fathers
	if (TaskGetLevel(gdbP, gActualTask) > 1)
	{
		TaskCalcCompleted(gdbP, 
			TaskGetFatherIndex(gdbP, gActualTask));
	}
	FrmUpdateForm(FrmGetActiveFormID(), frmRedrawUpdateCode);
}



/****************************************************************************
 * Name : PriorityHandleEvent
 * Desc : update the priority of the actualtask
 * Parm :
 * Out  : 
 * Auth : burgbach, 19.10.2000
 ***************************************************************************/
void PriorityHandleEvent(EventPtr e)
{
	Coord lx, ly;
	// edit it in place
	Int16 TInt16;
	ListPtr listP = GetObjectPtr(Lst_Priority);
	TaskFormatType format;

	format = TaskGetFormat(gdbP, gActualTask);
	if (format.itemType == linkType)
	{
		return;
	}

	// init lx and ly depending on the active form
	if (FrmGetActiveFormID() == FrmFlat)
	{
		lx = 25;
		ly = 15 + e->data.tblSelect.row * 11;
	}
	else
	{
		lx = 4 + (1 + TaskGetLevel(gdbP, gActualTask) - 
			gRefLevel) * 11;
		if (TaskGetType(gdbP, gActualTask) == informativeType)
			lx -= 11;
		ly = 15 + e->data.tblSelect.row * 11;
	}

	LstSetSelection(listP, 
		TaskGetPriority(gdbP, gActualTask) - 1);
	ly -= (TaskGetPriority(gdbP, gActualTask) - 1) * 11;
	// don't get out the screen
	ly = ly < 1 ? 1 : ly;
	ly = ly > 103 ? 103 : ly;
	LstSetPosition(listP, lx, ly);
	TInt16 = LstPopupList(listP);
	if (TInt16 != -1)
	{
		TaskSetPriority(gdbP, gActualTask, TInt16 + 1);
		FrmUpdateForm(FrmGetActiveFormID(), frmRedrawUpdateCode);
	}
	if (TaskGetFormat(gdbP, gActualTask).hasToDo)
	{
		TaskSetToDoStatus(gdbP, gActualTask, toDoPriority,
			0, TaskGetPriority(gdbP, gActualTask),
			gNoDate, 0);
	}
} // void PriorityHandleEvent(EventPtr e)



/****************************************************************************
 * Name : SelectRange
 * Desc : popup a user form to ask for subview or all
 * Parm :
 * Out  : Btn_Sub_All or Btn_Sub_Sub or Btn_Sub_Cancel
 * Auth : burgbach, 21.10.2000
 ***************************************************************************/
UInt16 SelectRange(void)
{
	UInt16 temp = Btn_Sub_All;

	if (gParentTask) // if actual subview is not the total view
	{
		FormPtr frmSub = FrmInitForm(FrmSubOrAll);
		temp = FrmDoDialog(frmSub);
		FrmDeleteForm(frmSub);
	}

	return temp;
} // UInt16 SelectRange(void)



/****************************************************************************
 * Name : FrmMainMenuHandleEvent
 * Desc : Main Menu event handler
 * Parm :   -> Event to handle
 * Out  :   <- true if event was handled
 * Auth : burgbach, ??.??.2000
 * Mod  : hcc, 28.04.2001
 *          Add Collapse Father command
 * Mod	: lb, 2001-09-06
 * 		  	adapt to 0.23 db format
 ***************************************************************************/
Boolean FrmMainMenuHandleEvent(EventPtr e)
{
	FormPtr frm;

	frm = FrmGetActiveForm();

	switch(e->data.menu.itemID) {
		case MenuMainNewTask:
			ViewCreateNewTask();
			break;
		case MenuMainTop:
			// if there aren't any records, then just ignore
			if (PgNumRecords(gdbP) <= 1)
				break;

			gActualTask = gParentTask + 1;
			SelectActualTask(true);
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			break;
		case MenuMainBottom:
			if (gRefLevel == 0)
				gActualTask = PgNumRecords(gdbP) - 1;
			else
				gActualTask = GetLastInSub(gParentTask);
			SelectActualTask(true);
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			break;
		case MenuMainRepair:
			TaskRepairTree(gdbP, 0, 0);
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			break;
		case MenuMainReInit:
			switch (SelectRange())
			{
				case Btn_Sub_All:
					if (ConfirmCustom(StrActionDelete, StrAllTasks) == 0)
					{
						while (PgNumRecords(gdbP) > 1)
						{
							TaskRemove(gdbP, 1);
						}
						gParentTask = gRefLevel = 0;
						// erase the up button
						CtlHideControl(GetObjectPtr(Btn_Sub_Up));
						FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
					}
					break;
				case Btn_Sub_Sub:
					if (ConfirmCustom(StrActionDelete, 
						StrAllDoneTasksInThisView) == 0)
					{
						TaskRemoveChildren(gdbP, gParentTask);
						FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
					}
					break;
				default:
					break;
			}
			break;
		case MenuMainSyncToDo:
				TaskSyncAll();
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			break;
		case MenuMainResetDates:
			FrmPopupForm(FrmSetReset);
			break;
		case MenuMainDeleteChildren:
			if (gActualTask && TaskGetHasChild(gdbP, gActualTask))
			{
				if (ConfirmCustom(StrActionDelete, 
					StrAllChildrenOfSelectedTask) == 0)
				{
					TaskRemoveChildren(gdbP, gActualTask);
					FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
				}
			}
			break;
		case MenuMainDeleteTask:
			if (QueryDeleteTask(gdbP,gActualTask))
			{
				TaskDeleteTree(gdbP,gActualTask);
				gActualTask=0;
				gParentTask=0;
				gRefLevel=0;
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			}	
			break;
		case MenuMainDeleteDoneTask:
			switch (SelectRange())
			{
				case Btn_Sub_All:
					if (ConfirmCustom(StrActionDelete, StrAllDoneTasks) == 0)
					{
						// to be sure the percentages are correct
						TaskCalcCompletedAll(gdbP);
						TaskRemoveDone(gdbP);
						TaskCalcCompletedAll(gdbP);
						FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
					}
					break;
				case Btn_Sub_Sub:
					if (ConfirmCustom(StrActionDelete, 
						StrAllDoneTasksInThisView) == 0)
					{
						// to be sure the percentages are correct
						TaskCalcCompletedAll(gdbP);
						TaskRemoveDoneChildren(gdbP, gParentTask);
						TaskCalcCompletedAll(gdbP);
						FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
					}
					break;
				default:
					break;
			}
			break;
		case MenuMainExpandAll:
			TaskExpand(gdbP, 0);
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			break;
		case MenuMainCollapseAll:
			TaskCollapse(gdbP, 0);
			Scroll(scrollTop, 0);
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			break;
//		case MenuMainExpandOne:
//			TaskExpand(gdbP, gActualTask);
//			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
//			break;
//		case MenuMainCollapseOne:
//			TaskCollapse(gdbP, gActualTask);
//			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
//			break;
		case MenuMainCollapseFather:
		{
			UInt16 fatherIndex = TaskGetFatherIndex(gdbP, gActualTask);
			if(fatherIndex != 0)
			{
				TaskCollapse(gdbP, fatherIndex);
				gActualTask = fatherIndex;
				SelectActualTask(true);
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			}
			break;
		}
		case MenuMainPreferences:
			ShowProjectPropertiesDialog();
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			break;
		case MenuMainDisplayPreferences:
			FrmPopupForm(FrmDisplayProperties);
			break;
		case MenuMainFlatFilter:
			FrmPopupForm(FrmFlatFilter);
			break;
		case MenuMainClose:
			FrmGotoForm(FrmProjectList);
			CloseDB(gdbP);
			gActualTask = gLastSelected = 0;
			break;
		case MenuMainViewSub:
			break;
		case MenuMainViewAll:
			gParentTask = 0;
			gRefLevel = 0;
			// erase the up button
			CtlHideControl(GetObjectPtr(Btn_Sub_Up));
			SetFormTitle(frm, gCurrentPrefs.openDBName);
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			break;
		case MenuEditCut:
			if (gActualTask)
			{
				TaskToClipboard(gdbP, gActualTask, true);
				TaskRemove(gdbP, gActualTask);
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			}
			break;
		case MenuEditCopy:
			if (gActualTask)
			{
				TaskToClipboard(gdbP, gActualTask, true);
			}
			break;
		case MenuEditCopyChildren:
			if (gActualTask && TaskGetHasChild(gdbP, gActualTask))
			{
				TaskToClipboard(gdbP, gActualTask, false);
			}
			break;
		case MenuEditPaste:
			if (gActualTask || PgNumRecords(gdbP) == 1 ||
				!TaskGetHasChild(gdbP, gParentTask))
			{
				if (!gActualTask)
					gActualTask = TaskFromClipboard(gdbP, gParentTask, true);
				else
					gActualTask = TaskFromClipboard(gdbP, gActualTask, false);
				SelectActualTask(true);
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			}
			break;
		case MenuEditPasteAsChildren:
			if (gActualTask)
				gActualTask = TaskFromClipboard(gdbP, gActualTask, true);
			SelectActualTask(true);
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			break;
		case MenuEditAddNote:
			if (gActualTask)
			{
				FrmPopupForm(FrmNoteEdit);
			}
			break;
		case MenuMainTaskDefaults:
			FrmPopupForm(FrmTaskDefaults);
			break;
		case MenuProjectListAbout:
			FrmCustomAlert(ProgectAbout, VERSION, " ", " ");
			break;
		case MenuMainMemoExport:
			FrmPopupForm(FrmMemoExportProperties);
			break;
		case MenuPublishLinkMaster:
			if (gActualTask > 0)
				PublishToLinkMaster(gdbP, gActualTask);
			break;
		case MenuLinkViaLinkMaster:
			if (gActualTask > 0)
				RequestLinkViaLinkMaster(gCurrentPrefs.openDBName, gActualTask);
			break;
		case MenuRemoveLinkViaLinkMaster:
			if (gActualTask > 0)
			{
				RemoveLinkViaLinkMaster(gActualTask);
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			}
			break;
	}
	return true;
} // Boolean FrmMainMenuHandleEvent(EventPtr e)



Boolean FrmMainCtlSelectHandleEvent(EventPtr e)
{
	FormPtr frm;
	
	frm = FrmGetActiveForm();

	switch(e->data.ctlSelect.controlID) {
		case Btn_Left:
			if (gActualTask)
			{
				TaskToLeft(gdbP, gActualTask);
				// TODO verify how to bypass this one
				ProjectTableUpdate();
				SelectActualTask(true);
			}
			break;
		case Btn_Right:
			if (gActualTask)
			{
				TaskToRight(gdbP, gActualTask);
				SelectActualTask(true);
			}
			break;
		case Btn_Up:
			if (gActualTask)
			{
				TaskUp(gdbP, gActualTask);
				SelectActualTask(true);
			}
			break;
		case Btn_Down:
			if (gActualTask)
			{
				TaskDown(gdbP, gActualTask);
				SelectActualTask(true);
			}
			break;
		case Btn_Sub:
			// New, be able to see the sub of a terminal task
			if (gActualTask)
			{
				char* p;
				gParentTask = gActualTask;
				gRefLevel = TaskGetLevel(gdbP, gParentTask);
				Scroll(scrollTop, 0);
				gActualTask = 0;
				// show the up button
				CtlShowControl(GetObjectPtr(Btn_Sub_Up));
				// update the titlebar
				p = TaskGetDescription(gdbP, gParentTask);
				SetFormTitle(frm, p);
				MemPtrFree(p);
				if (TaskGetHasChild(gdbP, gParentTask))
				{
					gActualTask = gParentTask + 1;
					SelectActualTask(true);
				}
			}
			break;
		case Btn_Sub_Up:
			TblSetItemInt(gTable, 0, 0, gParentTask);
			gParentTask = TaskGetFatherIndex(gdbP, gParentTask);
			if (!gParentTask)
			{
				gRefLevel = 0;
			}
			else
			{
				gRefLevel = TaskGetLevel(gdbP, gParentTask);
			}

			//this is done by the next update
			//if (gActualTask)
			//{
				//ProjectTableUpdate();
				//SelectActualTask(false);
			//}
			// if we're at top level
			if (!gParentTask)
			{
				// erase the up button
				CtlHideControl(GetObjectPtr(Btn_Sub_Up));
				SetFormTitle(frm, gCurrentPrefs.openDBName);
			}
			else
			{
				char* p = TaskGetDescription(gdbP, gParentTask);
				SetFormTitle(frm, p);
				MemPtrFree(p);
			}
			break;
		case Btn_Done:
			FrmGotoForm(FrmProjectList);
			CloseDB(gdbP);
			gActualTask = gLastSelected = gRefLevel = gParentTask = 0;
			break;
		case Btn_Main_Hierarchical:
			break;
		case Btn_Main_Flat:
			FrmGotoForm(FrmFlat);
			break;
		default:
			break;
	}
	FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
	return true;
} // Boolean FrmMainCtlSelectHandleEvent(EventPtr e)



Boolean FrmMainKeyDownHandleEvent(EventPtr e)
{
	Boolean handled = false;
	switch (e->data.keyDown.chr)
	{
		case vchrPageDown:
			Scroll(scrollDown, 5);
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			handled = true;
			break;
		case vchrPageUp:
			Scroll(scrollUp, 5);
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			handled = true;
			break;
		case chrSpace:
			TaskToRight(gdbP, gActualTask);
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			handled = true;
			break;
		case chrBackspace:
			TaskToLeft(gdbP, gActualTask);
			SelectActualTask(true);
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			handled = true;
			break;
		case chrLeftArrow:
			if (gActualTask)
			{
				TaskUp(gdbP, gActualTask);
				SelectActualTask(true);
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			}
			handled = true;
			break;
		case chrRightArrow:
			if (gActualTask)
			{
				TaskDown(gdbP, gActualTask);
				SelectActualTask(true);
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			}
			handled = true;
			break;
		case chrUpArrow:
		case vchrPrevField: // handera support
			if (gActualTask > gParentTask + 1)
			{
				UInt16 prev = gActualTask - 1;
				while (prev > gParentTask && !TaskIsVisible(gdbP, prev))
					prev--;
				if (prev > gParentTask)
				{
					gActualTask = prev;
					SelectActualTask(true);
					FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
				}
			}
			handled = true;
			break;
		case chrDownArrow:
		case vchrNextField: // handera support
			if ((gActualTask + 1 < PgNumRecords(gdbP)) &&
				TaskGetLevel(gdbP, gActualTask + 1) > gRefLevel)
			{
				UInt16 next = gActualTask + 1, numRec = PgNumRecords(gdbP);
				while (next < numRec && !TaskIsVisible(gdbP, next))
					next++;
				if (next < numRec)
				{
					gActualTask = next;
					SelectActualTask(true);
					FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
				}
			}
			handled = true;
			break;

//		case vchrHard3:
//			FrmGotoForm(FrmFlat);
//			break;

		case chrPlusSign:
			if (gActualTask)
			{
				TaskSetOpened(gdbP, gActualTask, true);
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			}
			handled = true;
			break;
		case chrHyphenMinus:
			if (gActualTask && TaskGetHasChild(gdbP, gActualTask))
			{
				TaskSetOpened(gdbP, gActualTask, false);
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
			}
			handled = true;
			break;

		default:
			// keydown is tab or return, create a new subtask 
			// (space is treated already)
			if (TxtGlueCharIsSpace(e->data.keyDown.chr))
			{
				if (gActualTask)
				{
					if (e->data.keyDown.chr == chrLineFeed)
					{
						AddRecordSub(gdbP, &gEmptyTask, &gActualTask);
						// recalc fathers progressbar
						if (TaskGetLevel(gdbP, gActualTask) > 1)
						{
							TaskCalcCompleted(gdbP, 
								TaskGetFatherIndex(gdbP, gActualTask));
						}
						FrmPopupForm(FrmTaskEdit);
						ProjectTableUpdate(); // prevent unnecessary scroll up
						SelectActualTask(true);
						FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
						handled = true;
					} 
					else if (e->data.keyDown.chr == chrHorizontalTabulation ||
							 e->data.keyDown.chr == chrCarriageReturn)
					{
						// edit the task (keyboard shortcut or jogdial)
						FrmPopupForm(FrmTaskEdit);
						FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
						handled = true;
					}
				}
			}
#if 1
// follow codes not support FEP
			// keyDown is alnum, then create a new task
			else if (TxtGlueCharIsAlNum(e->data.keyDown.chr))
			{
				if (gActualTask || PgNumRecords(gdbP) == 1 ||
					!TaskGetHasChild(gdbP, gParentTask))
				{
					TaskType NewTask = gEmptyTask;
					Char firstChar[2];
					firstChar[0] = 
						(Char)TxtGlueUpperChar(e->data.keyDown.chr);
					firstChar[1] = '\0';
					NewTask.description = firstChar;

					// special case when there's no existing task
					if (PgNumRecords(gdbP) == 1 || 
						!TaskGetHasChild(gdbP, gParentTask))
					{
						gActualTask = gParentTask;
					}
					AddRecord(gdbP, &NewTask, &gActualTask);

					// recalc fathers progressbar
					if (TaskGetLevel(gdbP, gActualTask) > 1)
					{
						TaskCalcCompleted(gdbP, 
							TaskGetFatherIndex(gdbP, gActualTask));
					}
					FrmPopupForm(FrmTaskEdit);
					ProjectTableUpdate(); // prevent unnecessary scroll up
					SelectActualTask(true);
					FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
					handled = true;
				}
			}
#else
			else if (TxtCharIsPrint (e->data.keyDown.chr))
			{
					ViewCreateNewTask();
					e->data.keyDown.chr = TxtGlueUpperChar(e->data.keyDown.chr);
					EvtAddEventToQueue(e);
					handled = true;
			}
#endif
			break;
	}
	return handled;
} // Boolean FrmMainKeyDownHandleEvent(EventPtr e)



static Boolean FrmMainHandleEvent (EventPtr e)
{
	Boolean handled = false;
	FormPtr frm;
	Int16 x;
	TaskFormatType format;
	
	frm = FrmGetActiveForm();
	gTable = GetObjectPtr(ProjectTable);

	switch (e->eType) {
	case frmUpdateEvent:
		switch(e->data.frmUpdate.updateCode)
		{
			case frmRedrawUpdateCode:
				ProjectTableUpdate();
				FrmDrawForm(frm);
				SelectActualTask(false);
				handled = true;
				break;
			default:
				break;
		}
		break;

	case frmOpenEvent:
		DBGMSG((DBB, "In FrmMainHandleEvent, frmOpenEvent"));
		if (OSCaps.handera)
		{
			VgaFormModify(frm, vgaFormModify160To240);
			VgaTableUseBaseFont(gTable, true);
		}
		gCurrentPrefs.view = normalView;
		ProjectTableInit();
		ProjectTableUpdate();
		SetFormTitle(frm, gCurrentPrefs.openDBName);
		FrmDrawForm(frm);
		SelectActualTask(true); // show on screen.
		gTable = GetObjectPtr(ProjectTable);
		CtlSetValue(GetObjectPtr(Btn_Main_Hierarchical), true);
		if (gRefLevel)
		{
			Char *p;
			// show the up button
			CtlShowControl(GetObjectPtr(Btn_Sub_Up));
			// update the titlebar
			p = TaskGetDescription(gdbP, gParentTask);
			// uh oh, sanity check failed, try to reset to defaults
			if (!p)
			{
				gActualTask=0;
				gParentTask=0;
				gRefLevel=0;
				break;
			}
			SetFormTitle(frm, p);
			MemPtrFree(p);
			if (TaskGetHasChild(gdbP, gParentTask))
			{
				gActualTask = gParentTask + 1;
				SelectActualTask(true);
			}
		}
		handled = true;
		break;

// TODO : Handera display extent, fix it
#if 0
	case displayExtentChangedEvent:
	{
		RectangleType win, tab;
		Coord x, y;
		Int8 direction;
		win.topLeft.x = 0;
		win.topLeft.y = 0;
		WinGetDisplayExtent(&win.extent.x, &win.extent.y);
		WinSetWindowBounds(FrmGetWindowHandle(frm), &win);
		TblGetBounds(gTable, &tab);
		tab.extent.y = win.extent.y;
		TblSetBounds(gTable, &tab);

		FrmHideObject(frm, Btn_Done);
		FrmHideObject(frm, Btn_Left);
		FrmHideObject(frm, Btn_Down);
		FrmHideObject(frm, Btn_Up);
		FrmHideObject(frm, Btn_Right);
		FrmHideObject(frm, Btn_Sub);
		FrmHideObject(frm, Btn_Sub_Up);
		FrmHideObject(frm, Btn_Scroll_Up);
		FrmHideObject(frm, Btn_Scroll_Down);

		direction = win.extent.y > 240 ? -80 : 80;
		
		FrmGetObjectPosition(frm, Btn_Done, &x, &y);
		FrmSetObjectPosition(frm, Btn_Done, x, y + direction);
		FrmGetObjectPosition(frm, Btn_Left, &x, &y);
		FrmSetObjectPosition(frm, Btn_Left, x, y + direction);
		FrmGetObjectPosition(frm, Btn_Down, &x, &y);
		FrmSetObjectPosition(frm, Btn_Down, x, y + direction);
		FrmGetObjectPosition(frm, Btn_Up, &x, &y);
		FrmSetObjectPosition(frm, Btn_Up, x, y + direction);
		FrmGetObjectPosition(frm, Btn_Right, &x, &y);
		FrmSetObjectPosition(frm, Btn_Right, x, y + direction);
		FrmGetObjectPosition(frm, Btn_Sub, &x, &y);
		FrmSetObjectPosition(frm, Btn_Sub, x, y + direction);
		FrmGetObjectPosition(frm, Btn_Sub_Up, &x, &y);
		FrmSetObjectPosition(frm, Btn_Sub_Up, x, y + direction);
		FrmGetObjectPosition(frm, Btn_Scroll_Up, &x, &y);
		FrmSetObjectPosition(frm, Btn_Scroll_Up, x, y + direction);
		FrmGetObjectPosition(frm, Btn_Scroll_Down, &x, &y);
		FrmSetObjectPosition(frm, Btn_Scroll_Down, x, y + direction);

		FrmShowObject(frm, Btn_Done);
		FrmShowObject(frm, Btn_Left);
		FrmShowObject(frm, Btn_Down);
		FrmShowObject(frm, Btn_Up);
		FrmShowObject(frm, Btn_Right);
		FrmShowObject(frm, Btn_Sub);
		FrmShowObject(frm, Btn_Sub_Up);
		FrmShowObject(frm, Btn_Scroll_Up);
		FrmShowObject(frm, Btn_Scroll_Down);

		FrmDrawForm(frm);
		handled = true;
		break;
	}
#endif

	case menuEvent:
		MenuEraseStatus(NULL);
		handled = FrmMainMenuHandleEvent(e);
		break;

	// remember the screen coords
	case tblEnterEvent:
		gLastDownX = e->screenX;
		gLastDownY = e->screenY;
		// deselect all
		gActualTask = 0;
		break;

	case ctlSelectEvent:
		handled = FrmMainCtlSelectHandleEvent(e);
		break;

	case ctlRepeatEvent:
		switch (e->data.ctlRepeat.controlID)
		{
			case Btn_Scroll_Up:
				Scroll(scrollUp, 1);
				break;
			case Btn_Scroll_Down:
				Scroll(scrollDown, 1);
				break;
			default:
				break;
		}
		// not handled, or I don't get the repeated event
		//handled = true;
//		FrmUpdateForm(FrmMain, wUpdateCode);
		FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
		break;

	case tblSelectEvent:
		gActualTask = TblGetItemInt(gTable, e->data.tblSelect.row, 0);
		if (gActualTask == -1)
			gActualTask = 0;

		format = TaskGetFormat(gdbP, gActualTask);

		x = gLastDownX / 11 + 1 - TaskGetLevel(gdbP, gActualTask) + gRefLevel;
		if (gActualTask) // this is for all the switch
		switch (x)
		{
			case 0:
				/**************************************************************
				 * Code for open buttons
				 *************************************************************/
				TaskToggleIsOpened(gdbP, gActualTask);
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
				break;
			case 1:
				/**************************************************************
				 * Code for progress buttons
				 *************************************************************/
				if (format.itemType == actionType || 
					format.itemType == progressType)
				{
					ProgressHandleEvent();
					break;
				}
				/* else FALLTHRU */
			default:
				//////////////////////////////////////////////////////////////
				// Code for note or follow link
				//////////////////////////////////////////////////////////////
				if (gLastDownX > (OSCaps.screenWidth - 7) && format.hasNote)
				{
					FrmPopupForm(FrmNoteEdit);
				}
				else if (gLastDownX > ((OSCaps.screenWidth - 7) 
					- (format.hasNote? 8 : 0))
					&& format.hasLink)
				{
					FollowLinkViaLinkMaster(gActualTask);
				}
				// if a priority is visible
				else if (x == 2 && 
					format.itemType != informativeType && 
					TaskGetPriority(gdbP, gActualTask) != NO_PRIORITY && 
					gProjectPrefs.displayPriorities)
				{
					PriorityHandleEvent(e);
				}
				else if (x == 1 && 
					format.itemType == informativeType && 
					TaskGetPriority(gdbP, gActualTask) != NO_PRIORITY &&
					gProjectPrefs.displayPriorities)
				{
					PriorityHandleEvent(e);
				}
				// edit text if click on the right side
				else if (x > 0)
				{
					if (gLastSelected == gActualTask)
					{
						if (format.itemType == linkType)
						{
		//					/// +++ FIX THIS +++
		//					Need to bring up a dialog to tell the user wo don't support anymore
		//					DirectLinkFollow(gdbP, gActualTask);
						}
						else
						{
							FrmPopupForm(FrmTaskEdit);
						}
					}
					else
					{
						gLastSelected = gActualTask;
					}
				}
				else
				{
					// left part
					// nothing to do for now
				}
				break;
		}
		handled = true;
		break;

	// TODO : put this '5' in prefs (scroll step)
	case keyDownEvent:
		handled = FrmMainKeyDownHandleEvent(e);
		break;

	// For multi-line items, the table cell is > 1 line high,
    // but PalmOS will not generate a tblSelectEvent if the user taps
	// below the first 11 pixels of the table row. This code manually
	// selects an item no matter where in the cell the tap occurs.
	case penDownEvent:
	{
		Int16 row;
		DBGMSG((DBB,"FrmMain penDown Event %d, penDown=%d, tapCount=%d, x=%d, y=%d\n",
				e->eType, (UInt16)e->penDown, (UInt16)e->tapCount,
				e->screenX, e->screenY));

		row = TblGetRowContainingPoint(gTable, e->screenX, e->screenY);

		if(row != -1)
		{
			gActualTask = TblGetItemInt(gTable, row, 0);
			TblSelectItem(gTable, row, 0);
		}
		break;
	}
	default:
		DBGMSG((DBB,"FrmMain Other Event %d, penDown=%d, tapCount=%d, x=%d, y=%d\n",
				e->eType, (UInt16)e->penDown, (UInt16)e->tapCount,
				e->screenX, e->screenY));
		break;
	}

	return handled;
} // static Boolean FrmMainHandleEvent (EventPtr e)



static void UpdateIconButton(void)
{
	FormPtr frm = FrmGetActiveForm();
	UInt16 iconid;
	UInt16 size = sizeof(iconid);
	if (!TaskGetExtraChunk(gdbP, gActualTask, Extra_Icon, 0,
								  &iconid, &size) &&
								  size == sizeof(iconid) &&
								  iconid < getNumIconsLoaded() )
	{
		Coord x, y;
		CtlSetLabel(GetObjectPtr(Btn_SelectIcon), " ");
		FrmGetObjectPosition(frm, 
			FrmGetObjectIndex(frm, Btn_SelectIcon),
			&x, &y);
		drawIcon(iconid, x + 10, y + 2);
	}
	else
	{
		MemHandle h;
		char* p;

		h = DmGetResource('tSTR', StrIconButton);
		p = MemHandleLock(h);

		CtlSetLabel(GetObjectPtr(Btn_SelectIcon), p);

		MemHandleUnlock(h);
		DmReleaseResource(h);
	}
}



static void FrmTaskEditPart(PartType part)
{
	FormPtr frm = FrmGetActiveForm();
	switch (part)
	{
		case showValue:
			FrmShowObject(frm, 
				FrmGetObjectIndex(frm, TaskNumericLimit));
			FrmShowObject(frm, 
				FrmGetObjectIndex(frm, TaskNumericActual));
			FrmShowObject(frm, 
				FrmGetObjectIndex(frm, LabelValue));
			FrmShowObject(frm, 
				FrmGetObjectIndex(frm, LabelSlash));
			break;
		case hideValue:
			FrmHideObject(frm, 
				FrmGetObjectIndex(frm, TaskNumericLimit));
			FrmHideObject(frm, 
				FrmGetObjectIndex(frm, TaskNumericActual));
			FrmHideObject(frm, 
				FrmGetObjectIndex(frm, LabelValue));
			FrmHideObject(frm, 
				FrmGetObjectIndex(frm, LabelSlash));
			break;
	}
} // static void FrmTaskEditPart(UInt16 part)



static Boolean FrmTaskEditHandleEvent (EventPtr e)
{
	Boolean handled = false;
	FormPtr frm;
	DateType date;
	
	frm = FrmGetActiveForm();

	switch (e->eType) {
	case frmOpenEvent:
		if (OSCaps.handera)
		{
			VgaFormModify(frm, vgaFormModify160To240);
		}
		DBGMSG((DBB, "Editing task %hu", gActualTask));
		DBGMSG((DBB, "Task at opening"));
		dbgPrintTask(gdbP, gActualTask);
		TaskSave(gActualTask);
		DBGMSG((DBB, "After having saved it"));
		dbgPrintTask(gdbP, gActualTask);
		InitFrmTaskEdit();
		DBGMSG((DBB, "After InitFrmTaskEdit"));
		dbgPrintTask(gdbP, gActualTask);
		FrmDrawForm(frm);
		UpdateIconButton();
		DBGMSG((DBB, "After UpdateIconButton"));
		dbgPrintTask(gdbP, gActualTask);
		FrmSetFocus(frm, FrmGetObjectIndex(frm, TaskDesc));
		handled = true;
		break;
	
	case frmCloseEvent:
		DBGMSG((DBB, "frmCloseEvent"));
		// if (false) we have no more description, we have to try to delete
		if (CleanUpFrmTaskEdit())
			{
				// link the todo
				if (CtlGetValue(GetObjectPtr(Chk_Edit_ToDo_Link)))
				{
					TaskPublishToDo(gdbP, gActualTask);
				}
				else
				{
					TaskRemoveHasToDo(gdbP, gActualTask);
				}
				TaskUpdateToDo(gdbP, gActualTask);
				TaskCalcCompleted(gdbP, gActualTask);
				TaskRemoveSaved(); 
			}
			else
			{
				if(!TaskGetHasChild(gdbP, gActualTask))
				{
					// No children, then just delete like
					// normal.
					TaskRemoveSaved(); 
					TaskDeleteTree(gdbP,gActualTask);
					gActualTask=0;
					gParentTask=0;
					gRefLevel=0;
				}
				else
				{
					// Because the task had children, we
					// won't presume to delete everything
					// for the sake of the user.
					TaskGetSaved(gActualTask);
				}
			}
		break;

	case frmUpdateEvent:
		switch(e->data.frmUpdate.updateCode)
		{
			case frmTaskEditReturnFromIconSelect:
				FrmDrawForm(frm);
				UpdateIconButton();
			case frmTaskEditReturnFromNote:
				FrmSetFocus(frm, FrmGetObjectIndex(frm, TaskDesc));
				handled = true;
				break;
		}
		break;

	case menuEvent:
		handled = EditMenuHandleEvent(e->data.menu.itemID, 
			GetObjectPtr(TaskDesc));
		break;

	case ctlSelectEvent:
		switch(e->data.ctlSelect.controlID) {
			case Pop_Due_Date:
				switch (LstPopupList(GetObjectPtr(Lst_Date)))
				{
					case 0: // today
						TaskSetDueDate(gdbP, gActualTask, gToday);
						break;
					case 1: // tomorrow
						date = gToday;
						DateAdjust(&date, 1);
						TaskSetDueDate(gdbP, gActualTask, date);
						break;
					case 2: // in one week
						date = gToday;
						DateAdjust(&date, 7);
						TaskSetDueDate(gdbP, gActualTask, date);
						break;
					case 3: // no date
						TaskSetDueDate(gdbP, gActualTask, gNoDate);
						break;
					case 4: // choose date
						TaskInputDueDate(gdbP, gActualTask);
						break;
					default:
						break;
				}
				date = TaskGetDueDate(gdbP, gActualTask);
				FrmTaskEditUpdateDateTrigger(GetObjectPtr(Pop_Due_Date), date);
				handled = true;
				break;

			case Btn_Plus_One:   /*FALLTHRU*/
			case Btn_Plus_Seven: /*FALLTHRU*/
			case Btn_Minus_One:
				date = TaskGetDueDate(gdbP, gActualTask);

				// init date if needed
				if (date.month == 0)
					date = gToday;

				switch (e->data.ctlSelect.controlID - Btn_Plus_Seven)
				{
					case 0 : // +7
						DateAdjust(&date, 7);
						break;
					case 1 : // -1
						DateAdjust(&date, -1);
						break;
					case 2 : // +1
						DateAdjust(&date, 1);
						break;
					default:
						break;
				}
				TaskSetDueDate(gdbP, gActualTask, date);
				FrmTaskEditUpdateDateTrigger(GetObjectPtr(Pop_Due_Date), date);
				handled = true;
				break;

			case Pop_Item_Type:
				if (CtlGetValue(GetObjectPtr(Chk_Edit_ToDo_Link)))
				{
					// initialize item type popup trigger
					//LstSetSelection(GetObjectPtr(Lst_Item_Type), 2);
					LstSetSelection(GetObjectPtr(Lst_Item_Type), actionType);
					TaskSetType(gdbP, gActualTask, actionType);
				}
				else
				{
					Int16 choice = LstPopupList(GetObjectPtr(Lst_Item_Type));
					switch (choice)
					{
						case 0: // progress
							TaskSetType(gdbP, gActualTask, progressType);
							break;
						case 1: // numeric, is a progress type with extrablock
							TaskSetType(gdbP, gActualTask, numericType);
							break;
						case 2: // action
							TaskSetType(gdbP, gActualTask, actionType);
							break;
						case 3: // informative
							TaskSetType(gdbP, gActualTask, informativeType);
							break;
						default:
							break;
					}
					if (choice == 1)
					{
						FrmTaskEditPart(showValue);
					}
					else if (choice != -1)
					{
						FrmTaskEditPart(hideValue);
					}
				}
				UpdatePopupTrigger(GetObjectPtr(Pop_Item_Type),
					GetObjectPtr(Lst_Item_Type));
				handled = true;
				break;

			case Pop_Category:
				{
					UInt16 attr;
					UInt16 categ;
					Char* name = (Char*)CtlGetLabel(GetObjectPtr(Pop_Category));
//					UInt16 idEditStr =
//						(OSCaps.ver32? categoryDefaultEditCategoryString:
//						               StrEditCategories);
					DmRecordInfo(gdbP, gActualTask, &attr, NULL, NULL);
					categ = attr & dmRecAttrCategoryMask;
					CategorySelect(gdbP, frm, Pop_Category, Lst_Category, false,
								   &categ, name,
								   1, NULL); // this works
								   //1, idEditStr);
					attr &= ~dmRecAttrCategoryMask;
					attr |= categ & dmRecAttrCategoryMask;
					DmSetRecordInfo(gdbP, gActualTask, &attr, NULL);
					CtlSetLabel(GetObjectPtr(Pop_Category), name);
				}
				handled = true;
				break;

			case Btn_Edit_Note:
				CleanUpFrmTaskEdit();
				FrmPopupForm(FrmNoteEdit);
				handled = true;
				break;

			case Btn_Edit_Ok:
				DBGMSG((DBB, "Btn_Edit_OK"));
				dbgPrintTask(gdbP, gActualTask);
				// if (false) we have no more description, we have to try to delete
				if (CleanUpFrmTaskEdit())
				{
					// link the todo
					if (CtlGetValue(GetObjectPtr(Chk_Edit_ToDo_Link)))
					{
						TaskPublishToDo(gdbP, gActualTask);
					}
					else
					{
						TaskRemoveHasToDo(gdbP, gActualTask);
					}
					TaskUpdateToDo(gdbP, gActualTask);
					TaskCalcCompleted(gdbP, gActualTask);
					TaskRemoveSaved(); 
				}
				else
				{
					if(!TaskGetHasChild(gdbP, gActualTask))
					{
						TaskRemoveSaved(); 
						TaskDeleteTree(gdbP,gActualTask);
						gActualTask=0;
						gParentTask=0;
						gRefLevel=0;
					}
					else
					{
						if(QueryDeleteTask(gdbP, gActualTask))
						{
							TaskRemoveSaved();
							TaskDeleteTree(gdbP,gActualTask);
							gActualTask=0;
							gParentTask=0;
							gRefLevel=0;
						}
						else
						{
							TaskGetSaved(gActualTask);
						}
					}
				}
				FrmReturnToForm(0);
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
				handled = true;
				break;

			case Btn_Edit_Cancel:
				CleanUpFrmTaskEdit();
				TaskGetSaved(gActualTask);
				// +++ CHECK THIS +++
				// does this make sure the task gets deleted if it has no description?
				// +++ CHECK THIS +++
				if (TaskGetFormat(gdbP, gActualTask).newTask)
				{
					TaskRemove(gdbP, gActualTask);
				}
				FrmReturnToForm(0);
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
				gActualTask=0;
				handled = true;
				break;

			case Btn_SelectIcon:
				if (getNumIconsLoaded() != 0)
				{
					CleanUpFrmTaskEdit();
					FrmPopupForm(FrmIconSelect);
				}
				handled = true;
				break;

			case Chk_Edit_ToDo_Link:
				// real job is done by OK button
				LstSetSelection(GetObjectPtr(Lst_Item_Type), actionType);
				UpdatePopupTrigger(GetObjectPtr(Pop_Item_Type),
					GetObjectPtr(Lst_Item_Type));
				TaskSetType(gdbP, gActualTask, actionType);
				FrmTaskEditPart(hideValue);
				handled = true;
				break;

			case Btn_Priority_1:
			case Btn_Priority_2:
			case Btn_Priority_3:
			case Btn_Priority_4:
			case Btn_Priority_5:
			case Btn_Priority_No:
				TaskSetPriority(gdbP, gActualTask, e->data.ctlSelect.controlID - Btn_Priority_X + 1);
				handled = true;

			default:
				break;
		}
		break;

	case keyDownEvent:
		switch (e->data.keyDown.chr)
		{
			// on return, close form
			case chrCarriageReturn:
			case chrLineFeed:
				dbgPrintTask(gdbP, gActualTask);
				// if (false) we have no more description, we have to try to delete
				if (CleanUpFrmTaskEdit())
				{
					// link the todo
					if (CtlGetValue(GetObjectPtr(Chk_Edit_ToDo_Link)))
					{
						TaskPublishToDo(gdbP, gActualTask);
					}
					else
					{
						TaskRemoveHasToDo(gdbP, gActualTask);
					}
					TaskUpdateToDo(gdbP, gActualTask);
					TaskCalcCompleted(gdbP, gActualTask);
					TaskRemoveSaved(); 
				}
				else
				{
					if(!TaskGetHasChild(gdbP, gActualTask))
					{
						TaskRemoveSaved(); 
						TaskDeleteTree(gdbP,gActualTask);
						gActualTask=0;
						gParentTask=0;
						gRefLevel=0;
					}
					else
					{
						if(QueryDeleteTask(gdbP, gActualTask))
						{
							TaskRemoveSaved();
							TaskDeleteTree(gdbP,gActualTask);
							gActualTask=0;
							gParentTask=0;
							gRefLevel=0;
						}
						else
						{
							TaskGetSaved(gActualTask);
						}
					}
				}
				FrmReturnToForm(0);
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
				handled = true;
				break;
				{
					UInt8 priority;
					// get the priority and set it on the form
					priority = TaskGetPriority(gdbP, gActualTask);
					// old databases
					if (priority == 0)
						priority = NO_PRIORITY;
					CtlSetValue(GetObjectPtr(Btn_Priority_X + priority - 1), 
						false);
					priority++;
					if (priority > NO_PRIORITY)
						priority = 1;
					TaskSetPriority(gdbP, gActualTask, priority);
					CtlSetValue(GetObjectPtr(Btn_Priority_X + priority - 1), 
						true);
				}
				handled = true;
				break;
			default:
				break;
		}
		break;

	default:
		break;
	}

	return handled;
} // static Boolean FrmTaskEditHandleEvent (EventPtr e)


static Boolean FrmFlatFilterHandleEvent (EventPtr e)
{
	Boolean handled = false;
	FormPtr frm;
	Int16 TInt16;
	ListPtr listP = NULL;
	
	frm = FrmGetActiveForm();

	switch (e->eType) {
	case frmOpenEvent:
		if (OSCaps.handera)
		{
			VgaFormModify(frm, vgaFormModify160To240);
		}
		// set the controls based on the actual prefs
		CtlSetValue(GetObjectPtr(Chk_Flat_Hide), 
			gProjectPrefs.flatHideDone);
		CtlSetValue(GetObjectPtr(Btn_No_Date + gProjectPrefs.flatDated), true);
		CtlSetValue(GetObjectPtr(Btn_Min + !gProjectPrefs.flatMin), true);
		CtlSetValue(GetObjectPtr(
			Btn_Priority_X + gProjectPrefs.flatMinPriority - 1)
			, true);
		CtlSetValue(GetObjectPtr(Btn_And + gProjectPrefs.flatOr), true);
		CtlSetValue(GetObjectPtr(Btn_No_Sort + gProjectPrefs.flatSorted), true);

		LstSetSelection(GetObjectPtr(List_DaysToBold), 
			gProjectPrefs.flatDateLimit);

		CtlSetLabel(GetObjectPtr(Pop_DaysToBold),
			LstGetSelectionText(GetObjectPtr(List_DaysToBold),
			gProjectPrefs.flatDateLimit) );

		FrmDrawForm(frm);
		handled = true;
		break;
	
	case frmCloseEvent:
		// handled = true ??
		break;

	case menuEvent:
		handled = true;
		break;

	case ctlSelectEvent:
		switch(e->data.ctlSelect.controlID) {
			case Pop_DaysToBold:
				listP = GetObjectPtr(e->data.ctlSelect.controlID + 1);
				TInt16 = LstPopupList(listP);
				if (TInt16 != -1)
				{
					CtlSetLabel(GetObjectPtr(e->data.ctlSelect.controlID),
						LstGetSelectionText(listP, TInt16));
				}
				handled = true;
				break;
				
			case Btn_Prop_Ok:
				// modify the prefs
				gProjectPrefs.flatHideDone = 
					CtlGetValue(GetObjectPtr(Chk_Flat_Hide));
				gProjectPrefs.flatDated = 
					FrmGetObjectId(frm, FrmGetControlGroupSelection(frm, 3)) -
					Btn_No_Date;

				gProjectPrefs.flatMinPriority =
					FrmGetObjectId(frm, FrmGetControlGroupSelection(frm, 1)) -
					Btn_Priority_X + 1;

				gProjectPrefs.flatOr =
					FrmGetObjectId(frm, FrmGetControlGroupSelection(frm, 2)) -
					Btn_And;

				gProjectPrefs.flatMin =
					!(FrmGetObjectId(frm, FrmGetControlGroupSelection(frm, 4)) -
					Btn_Min);

				gProjectPrefs.flatSorted =
					FrmGetObjectId(frm, FrmGetControlGroupSelection(frm, 5)) -
					Btn_No_Sort;

				gProjectPrefs.flatDateLimit = 
					LstGetSelection(GetObjectPtr(List_DaysToBold));

				if (gCurrentPrefs.view == flatView)
				{
					// reconstruct the list
					FlatCreateList(gdbP);
					FrmUpdateForm(FrmFlat, frmRedrawUpdateCode);
				}
				// fall through
			case Btn_Prop_Cancel:
				// return to caller without modifying the prefs
				FrmReturnToForm(0);
				handled = true;
				break;
			case Btn_Categories:
				selectCategories(&gProjectPrefs.flatCategories);
				handled = true;
			default:
				break;
		}
		break;

	case keyDownEvent:
		break;

	default:
		break;
	}

	return handled;
} // static Boolean FrmFlatFilterHandleEvent (EventPtr e)

void SelectProjectCategory (void)
{
	FormType* frmP;
	UInt16 category;
	Boolean categoryEdited;
	char CategoryName[dmCategoryLength];

	ProjectListMODELInit();
// +++ FIX THIS +++
// Somehow this code displays garbage when it first starts up, I really don't know why yet
// +++ FIX THIS +++
//
	// Process the category popup list.
	category=ProjectListModelGetProjectCategory(gCardNo,gCurrentPrefs.openDBName+5);

	CategoryGetName (ProjectListMODELGetDBRef(), category, CategoryName);

	frmP = FrmGetActiveForm();
	categoryEdited = CategorySelect (ProjectListMODELGetDBRef(), frmP, ProjectPreferencesCategoryPopup,
		 ProjectPreferencesCategoryList, false, &category, CategoryName, 1, categoryDefaultEditCategoryString);

	if ( categoryEdited || (category != ProjectListMODELGetCurrentCategory()))
	{
		ProjectListModelSetProjectCategory(gCardNo,gCurrentPrefs.openDBName+5,category);
	}

	ProjectListMODELCleanUp();
}

// BEFORE YOU CALL THIS, make sure you have a database open...
UInt16 ShowProjectPropertiesDialog(void)
{
	FormPtr frm;
	UInt16 hitButton;
	UInt16 attr;
	ControlType *popup;
	char CategoryName[dmCategoryLength];
	
// +++ FIX THIS +++
// This code needs to not change the category when someone hits cancel, currently it still changes
// the category, that is bad...
// +++ FIX THIS +++

		frm=FrmInitForm(FrmProjectProperties);

		//
		// Setup Category List
		//
		ProjectListMODELInit();
		CategoryGetName (ProjectListMODELGetDBRef(), ProjectListModelGetProjectCategory(gCardNo,gCurrentPrefs.openDBName+5), CategoryName);
		popup = FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, ProjectPreferencesCategoryPopup));
		CategorySetTriggerLabel (popup, CategoryName);
		ProjectListMODELCleanUp();

		//
		// set the controls based on the actual prefs
		//
		CtlSetValue(GetObjectFromForm(frm,Chk_Prop_Hide), gProjectPrefs.hideDoneTasks);
		CtlSetValue(GetObjectFromForm(frm,Chk_Prop_Display_Due_Dates), gProjectPrefs.displayDueDates);
		CtlSetValue(GetObjectFromForm(frm,Chk_Prop_Display_Priorities), gProjectPrefs.displayPriorities);
		CtlSetValue(GetObjectFromForm(frm,Chk_Prop_Display_Year), gProjectPrefs.displayYear);
		CtlSetValue(GetObjectFromForm(frm,Chk_Prop_Delete_Warn), gSavedPrefs.deleteWarn);
		CtlSetValue(GetObjectFromForm(frm,Chk_Prop_Father_Status), gProjectPrefs.useFatherStatus);
		CtlSetValue(GetObjectFromForm(frm,Chk_Prop_Auto_Sync_ToDo), gProjectPrefs.autoSyncToDo);
		CtlSetValue(GetObjectFromForm(frm,Chk_Prop_Comp_Date), gProjectPrefs.completionDate);

		//
		// read database attr
		//
		DmDatabaseInfo(0, gdbID, NULL, &attr, NULL, NULL, NULL, NULL, NULL, 
			NULL, NULL, NULL, NULL);
		CtlSetValue(GetObjectFromForm(frm,Chk_Prop_Save_DB), attr & dmHdrAttrBackup);

		FrmDrawForm(frm); // have to do now to be compatible with pre-3.5

 		FrmSetEventHandler(frm,FrmProjectPropertiesHandleEvent);
	
		hitButton=FrmDoDialog(frm);
		switch(hitButton)
		{
			case Btn_Prop_Ok:
				// modify the prefs
				gProjectPrefs.hideDoneTasks = 
					CtlGetValue(GetObjectFromForm(frm,Chk_Prop_Hide));
				gProjectPrefs.displayDueDates = 
					CtlGetValue(GetObjectFromForm(frm,Chk_Prop_Display_Due_Dates));
				gProjectPrefs.displayPriorities = 
					CtlGetValue(GetObjectFromForm(frm,Chk_Prop_Display_Priorities));
				gProjectPrefs.displayYear = 
					CtlGetValue(GetObjectFromForm(frm,Chk_Prop_Display_Year));
				gProjectPrefs.useFatherStatus = 
					CtlGetValue(GetObjectFromForm(frm,Chk_Prop_Father_Status));
				gProjectPrefs.autoSyncToDo = 
					CtlGetValue(GetObjectFromForm(frm,Chk_Prop_Auto_Sync_ToDo));
				gSavedPrefs.deleteWarn = 
					CtlGetValue(GetObjectFromForm(frm,Chk_Prop_Delete_Warn));
				gProjectPrefs.completionDate = 
					CtlGetValue(GetObjectFromForm(frm,Chk_Prop_Comp_Date));
				// read database attr
				DmDatabaseInfo(0, gdbID, NULL, &attr, NULL, NULL, NULL, NULL, 
					NULL, NULL, NULL, NULL, NULL);
				attr = attr | CtlGetValue(GetObjectFromForm(frm,Chk_Prop_Save_DB)) ? 
					dmHdrAttrBackup : 0;
				DmSetDatabaseInfo(0, gdbID, NULL, &attr, NULL, NULL, NULL, 
					NULL, NULL, NULL, NULL, NULL, NULL);
				break;

			case Btn_Prop_Cancel:
				break;

		case Btn_Choose_Name_Cancel:
			break;
	}

	FrmDeleteForm(frm);

	return hitButton;
}

static Boolean FrmProjectPropertiesHandleEvent (EventPtr e)
{
	Boolean handled = false;
	FormPtr frm;
	
	frm = FrmGetActiveForm();

	switch (e->eType) {
	case frmOpenEvent:
		handled = true;
		break;
	
	case frmCloseEvent:
		// handled = false
		break;

	case menuEvent:
		handled = true;
		break;

	case ctlSelectEvent:
		switch(e->data.ctlSelect.controlID) {
			case ProjectPreferencesCategoryPopup:
				SelectProjectCategory();
				handled = true;
				break;
/*
			case Btn_Prop_Ok:
				handled = true;
				break;
			case Btn_Prop_Cancel:
				handled = true;
				break;
*/
			default:
				break;
		}
		break;

	case keyDownEvent:
		break;

	default:
		break;
	}

	return handled;
} // static Boolean FrmProjectPropertiesHandleEvent (EventPtr e)



static Boolean FrmMemoExportPropertiesHandleEvent (EventPtr e)
{
	Boolean handled = false;
	FormPtr frm;
	Char exportName[dmDBNameLength];
	
	frm = FrmGetActiveForm();

	switch (e->eType) {
	case frmOpenEvent:

		if (OSCaps.handera)
		{
			VgaFormModify(frm, vgaFormModify160To240);
		}
		// set the controls based on the actual prefs
		CtlSetValue(GetObjectFromForm(frm,Chk_Exp_Done), 
			gMemoExportPrefs.exportDone);
		CtlSetValue(GetObjectFromForm(frm,Chk_Exp_Due_Dates), 	
			gMemoExportPrefs.exportDueDate);
		CtlSetValue(GetObjectPtr(Chk_Exp_Priorities), 
			gMemoExportPrefs.exportPriority);
		CtlSetValue(GetObjectPtr(Chk_Exp_Progress), 
			gMemoExportPrefs.exportProgress);
		CtlSetValue(GetObjectPtr(Chk_Exp_Note), 
			gMemoExportPrefs.exportNote);
		CtlSetValue(GetObjectPtr(Chk_Exp_Flat), 
			gMemoExportPrefs.exportFlat);

		LstSetSelection(GetObjectPtr(Lst_DocFormat), 
			gMemoExportPrefs.exportDocFormat);
		CtlSetLabel(GetObjectPtr(Pop_DocFormat),
			LstGetSelectionText(GetObjectPtr(Lst_DocFormat),
			gMemoExportPrefs.exportDocFormat) );

		FrmDrawForm(frm);

		handled = true;
		break;
	
	case frmCloseEvent:
		// handled = false
		break;

	case menuEvent:
		handled = true;
		break;

	case ctlSelectEvent:
		switch(e->data.ctlSelect.controlID) {
			case Pop_DocFormat:
			{
				ListPtr lst = GetObjectPtr(Lst_DocFormat);
				Int16 sel = LstPopupList(lst);
				if (sel != -1)
					CtlSetLabel(GetObjectPtr(Pop_DocFormat),
								LstGetSelectionText(lst, sel));
				handled = true;
				break;
			}
			case Btn_ToMemo:
			case Btn_ToDoc:
				// modify the prefs
				gMemoExportPrefs.exportDone = 
					CtlGetValue(GetObjectPtr(Chk_Exp_Done));
				gMemoExportPrefs.exportDueDate = 
					CtlGetValue(GetObjectPtr(Chk_Exp_Due_Dates));
				gMemoExportPrefs.exportPriority = 
					CtlGetValue(GetObjectPtr(Chk_Exp_Priorities));
				gMemoExportPrefs.exportProgress = 
					CtlGetValue(GetObjectPtr(Chk_Exp_Progress));
				gMemoExportPrefs.exportNote = 
					CtlGetValue(GetObjectPtr(Chk_Exp_Note));
				gMemoExportPrefs.exportFlat = 
					CtlGetValue(GetObjectPtr(Chk_Exp_Flat));
				gMemoExportPrefs.exportDocFormat =
					LstGetSelection(GetObjectPtr(Lst_DocFormat));

				if (e->data.ctlSelect.controlID == Btn_ToMemo)
				{
					if (ExportToMemo(gdbP, gMemoExportPrefs))
						DEBUG1("Problem during export");
					else
						DEBUG1("Ok");
				}
				else
				{
					if (Btn_Choose_Name_Ok ==  ShowChooseNameDialog("",exportName))
					{
						switch (SelectRange())
						{
							case Btn_Sub_All:
								ExportToDoc(0, exportName, 1, PgNumRecords(gdbP) - 1);
								break;
							case Btn_Sub_Sub:
								ExportToDoc(0, exportName, gParentTask, GetLastInSub(gParentTask));
								break;
							default:
								break;
						}
					}
				}
				handled=true;
				break;

				// fall through
			case Btn_Cancel:
				FrmReturnToForm(0);
				handled = true;
				break;
			default:
				break;
		}
		break;

	default:
		break;
	}

	return handled;
} // static Boolean FrmMemoExportPropertiesHandleEvent (EventPtr e)



static Boolean FrmSetResetHandleEvent (EventPtr e)
{
	Boolean handled = false;
	FormPtr frm;
	UInt16 first, last, i, idRange;
	
	frm = FrmGetActiveForm();

	switch (e->eType) {
	case frmOpenEvent:
		if (OSCaps.handera)
		{
			VgaFormModify(frm, vgaFormModify160To240);
		}
		// set the controls based on the actual prefs
		CtlSetValue(GetObjectPtr(Chk_Reset_Dates), false);
		CtlSetValue(GetObjectPtr(Chk_Reset_Progress), false);
		CtlSetValue(GetObjectPtr(Chk_Reset_Priority), false);
		CtlSetValue(GetObjectPtr(Chk_ToDo_Unlink), false);
		CtlSetValue(GetObjectPtr(Chk_ToDo_Link), false);
		FrmDrawForm(frm);
		handled = true;
		break;
	
	case frmCloseEvent:
		// handled = true ??
		break;

	case menuEvent:
		handled = true;
		break;

	case ctlSelectEvent:
		switch(e->data.ctlSelect.controlID) {
			case Btn_Ok:
				switch (SelectRange())
				{
					case Btn_Sub_All:
						first = 1;
						last = PgNumRecords(gdbP);
						idRange = StrAllTasks;
						break;
					case Btn_Sub_Sub:
						first = gParentTask + 1;
						last = GetLastInSub(gParentTask) + 1;
						idRange = StrAllTasksInThisSubview;
						break;
					default:
						return true;
						break;
				}
				if (ConfirmCustom(StrActionModify, idRange))
					return true;
				// modify the prefs
				for (i = first; i < last; i++)
				{
					if (CtlGetValue(GetObjectPtr(Chk_Reset_Dates)))
						TaskSetDueDate(gdbP, i, gNoDate);

					if (CtlGetValue(GetObjectPtr(Chk_Reset_Progress)))
					{
						ItemType type = TaskGetType(gdbP, i);
						if (type == actionType || type == progressType)
						{
							TaskSetCompleted(gdbP, i, 0, true);
						}
					}

					if (CtlGetValue(GetObjectPtr(Chk_Reset_Priority)))
						TaskSetPriority(gdbP, i, NO_PRIORITY);

					if (CtlGetValue(GetObjectPtr(Chk_ToDo_Unlink)) &&
						TaskGetFormat(gdbP, i).hasToDo)
						TaskRemoveHasToDo(gdbP, i);

					if (CtlGetValue(GetObjectPtr(Chk_ToDo_Link)) &&
						!TaskGetHasChild(gdbP, i))
					{
						TaskSetType(gdbP, i, actionType);
						// don't link a completed action
						if (TaskGetCompleted(gdbP, i) == ACTION_NOT_DONE)
						{
							TaskPublishToDo(gdbP, i);
							if (TaskGetPriority(gdbP, i) == NO_PRIORITY)
								TaskSetPriority(gdbP, i, 5);
						}
					}

					TaskUpdateToDo(gdbP, i);
				}
				FrmUpdateForm(FrmMain, frmRedrawUpdateCode);

				// fall through
			case Btn_Cancel:
				FrmReturnToForm(0);
				handled = true;
				break;
			default:
				break;
		}
		break;

	default:
		break;
	}

	return handled;
} // static Boolean FrmSetResetHandleEvent (EventPtr e)





//
// Event handler for display property dialog.
//
static Boolean FrmDisplayPropertiesHandleEvent (EventPtr e)
{
	Int16 TInt16;
	ListPtr listP;
	Boolean handled = false;
	FormPtr frm;
	
	frm = FrmGetActiveForm();

	switch (e->eType) {
		case frmOpenEvent:
			if (OSCaps.handera)
			{
				VgaFormModify(frm, vgaFormModify160To240);
			}
			LstSetSelection(GetObjectPtr(List_PriorityToBold), 
				gProjectPrefs.boldMinPriority);

			CtlSetLabel(GetObjectPtr(Pop_PriorityToBold),
				LstGetSelectionText(GetObjectPtr(List_PriorityToBold),
				gProjectPrefs.boldMinPriority) );

			LstSetSelection(GetObjectPtr(List_DaysToBold), 
				gProjectPrefs.boldMinDays);

			CtlSetLabel(GetObjectPtr(Pop_DaysToBold),
				LstGetSelectionText(GetObjectPtr(List_DaysToBold),
				gProjectPrefs.boldMinDays) );

			LstSetSelection(GetObjectPtr(List_WordWrapLines), 
				gProjectPrefs.wordWrapLines - 1);

			CtlSetLabel(GetObjectPtr(Pop_WordWrapLines),
				LstGetSelectionText(GetObjectPtr(List_WordWrapLines),
				gProjectPrefs.wordWrapLines - 1));

			CtlSetValue(GetObjectPtr(Chk_DispProp_StrikeDoneTasks), 
				gProjectPrefs.strikeDoneTasks);

			CtlSetValue(GetObjectPtr(Chk_DispProp_HideDoneProgress), 
				gProjectPrefs.hideDoneProgress);

			CtlSetValue(GetObjectPtr(Chk_DispProp_HideAllProgress), 
				gProjectPrefs.hideProgress);

			CtlSetValue(GetObjectPtr(Chk_DispProp_DrawTreeLines), 
				gProjectPrefs.drawTreeLines);

			FrmDrawForm(FrmGetActiveForm());
			handled = true;
			break;

		case menuEvent:
			handled = true;

		case ctlSelectEvent:
			switch(e->data.ctlSelect.controlID) {
				case Pop_PriorityToBold:
				case Pop_DaysToBold:
				case Pop_WordWrapLines:
					
					listP = GetObjectPtr(e->data.ctlSelect.controlID + 1);
					TInt16 = LstPopupList(listP);
					if (TInt16 != -1)
					{
						CtlSetLabel(GetObjectPtr(e->data.ctlSelect.controlID),
							LstGetSelectionText(listP, TInt16));
					}
					handled = true;
					break;

				case Btn_Ok:
					gProjectPrefs.boldMinPriority = 
						LstGetSelection(GetObjectPtr(List_PriorityToBold));

					gProjectPrefs.boldMinDays = 
						LstGetSelection(GetObjectPtr(List_DaysToBold));

					gProjectPrefs.wordWrapLines = 
						LstGetSelection(GetObjectPtr(List_WordWrapLines)) + 1;

					gProjectPrefs.strikeDoneTasks = 
						CtlGetValue(GetObjectPtr(Chk_DispProp_StrikeDoneTasks));

					gProjectPrefs.hideDoneProgress = 
						CtlGetValue(GetObjectPtr(Chk_DispProp_HideDoneProgress));

					gProjectPrefs.hideProgress = 
						CtlGetValue(GetObjectPtr(Chk_DispProp_HideAllProgress));

					gProjectPrefs.drawTreeLines = 
						CtlGetValue(GetObjectPtr(Chk_DispProp_DrawTreeLines));

					FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
					// fall down

				case Btn_Cancel: // return to caller without modifying the prefs
					FrmReturnToForm(0);
					handled = true;
					break;
			}

		default:
			break;
		}

	return handled;
} // static Boolean FrmDisplayPropertiesHandleEvent (EventPtr e)






/* this function from oreilly's palm programming book */
Boolean EditMenuHandleEvent(UInt16 menuid, FieldPtr fld)
{
	switch (menuid) {
		case MenuEditUndo:
		case MenuEditCut:
		case MenuEditCopy:
		case MenuEditPaste:
		case MenuEditSelectAll:
			if (!fld)
				return false;

			if (menuid==MenuEditUndo)
				FldUndo(fld);
			else if (menuid==MenuEditCut)
				FldCut(fld);
			else if (menuid==MenuEditCopy)
				FldCopy(fld);
			else if (menuid==MenuEditPaste)
				FldPaste(fld);
			else if (menuid==MenuEditSelectAll)
				FldSetSelection(fld, 0, FldGetTextLength(fld));

			return true;

		case MenuEditKeyboard:
			SysKeyboardDialog(kbdDefault);
			return true;

		case MenuEditGraffiti:
			SysGraffitiReferenceDialog(referenceDefault);
			return true;
		
		default:
			return false;
	};
} // Boolean EditMenuHandleEvent(UInt16 menuid)



void FrmNoteEditUpdateScroll(void)
{
	UInt16 textHeight, fieldHeight, scrollPos, maxValue;
	FieldPtr fld = GetObjectPtr(TaskNote);
	ScrollBarPtr bar = GetObjectPtr(Scr_Note_Edit);

	FldGetScrollValues(fld, &scrollPos, &textHeight, &fieldHeight);

	if (textHeight > fieldHeight)
		maxValue = (textHeight - fieldHeight) + FldGetNumberOfBlankLines (fld);
	else if (scrollPos)
		maxValue = scrollPos;
	else
		maxValue = 0;

	SclSetScrollBar (bar, scrollPos, 0, maxValue, fieldHeight-1);
} // void FrmNoteEditUpdateScroll(void)



void FrmNoteEditScrollPage(WinDirectionType direction)
{
	UInt16 textHeight, fieldHeight, scrollPos, maxValue;
	FieldPtr fld = GetObjectPtr(TaskNote);
	ScrollBarPtr bar = GetObjectPtr(Scr_Note_Edit);
	UInt16 step = 10;

	FldGetScrollValues(fld, &scrollPos, &textHeight, &fieldHeight);
	switch (direction)
	{
		case winUp:
			if (scrollPos - step < 0)
			{
				step = scrollPos;
				scrollPos = 0;
			}
			else
				scrollPos -= step;
			break;
		case winDown:
			if (scrollPos + step > textHeight)
			{
				step = textHeight - scrollPos;
				scrollPos = textHeight;
			}
			else
				scrollPos += step;
			break;
		default:
			break;
	}

	FldScrollField(fld, step, direction);

	FldGetScrollValues(fld, &scrollPos, &textHeight, &fieldHeight);

	if (textHeight > fieldHeight)
		maxValue = textHeight - fieldHeight;
	else if (scrollPos)
		maxValue = scrollPos;
	else
		maxValue = 0;

	SclSetScrollBar (bar, scrollPos, 0, maxValue, 
		fieldHeight-1);
} // void FrmNoteEditScrollPage(void)



static Boolean FrmNoteEditHandleEvent (EventPtr e)
{
	Boolean handled = false;
	FormPtr frm;
	MemHandle h;
	TaskExtendedRecordType *p;

	frm = FrmGetActiveForm();

	switch (e->eType) {
	case frmOpenEvent:
		if (OSCaps.handera)
		{
			VgaFormModify(frm, vgaFormModify160To240);
		}
		InitFrmNoteEdit();
		h = DmQueryRecord(gdbP, gActualTask);
		p = MemHandleLock(h);
		SetFormTitle(frm, (Char*)&StdFields(p)->description);
		MemHandleUnlock(h);
		FrmDrawForm(frm);
		// set the focus
		FrmSetFocus(frm, FrmGetObjectIndex(frm, TaskNote));
		FrmNoteEditUpdateScroll();
		handled = true;
		break;
	
	case frmCloseEvent:
		CleanUpFrmNoteEdit(); // save the note !!!
		// handled = true ??
		break;

	case menuEvent:
		handled = EditMenuHandleEvent(e->data.menu.itemID, 
			GetObjectPtr(TaskNote));
		break;

	case ctlSelectEvent:
		switch(e->data.ctlSelect.controlID) {
			case Btn_Note_Ok:
				// save
				CleanUpFrmNoteEdit();
			case Btn_Note_Cancel:
				// return to caller
				FrmUpdateForm(FrmTaskEdit, frmTaskEditReturnFromNote);
				FrmReturnToForm(0);
				handled = true;
				break;
			case Btn_Note_Delete:
				if (ConfirmCustom(StrActionDelete, StrThisNote) == 0)
				{
					FieldPtr fld = GetObjectPtr(TaskNote);
					MemHandle textH = FldGetTextHandle(fld);
					Char* textP;

					textP = MemHandleLock(textH);
					*textP=0; // Essentially delete the text
					MemHandleUnlock(textH);
					FldSetTextHandle(fld,textH);
					FldCompactText(fld);
					// save
					CleanUpFrmNoteEdit();
					// return to caller
       					FrmUpdateForm(FrmTaskEdit, frmTaskEditReturnFromNote);
					FrmReturnToForm(0);
					handled = true;
				}
				break;
			case Btn_Date:
				{
					Char date[dowLongDateStrLength];

					DateToAscii(gToday.month, gToday.day, 
						gToday.year + YEAR_OFFSET, PrefGetPreference(prefDateFormat), date);
					FldInsert(GetObjectPtr(TaskNote), date, StrLen(date));
				}
				break;
				
			default:
				break;
		}
		break;

	case keyDownEvent:
		switch (e->data.keyDown.chr)
		{
			case vchrPageDown:
				FrmNoteEditScrollPage(winDown);
				handled = true;
				break;
			case vchrPageUp:
				FrmNoteEditScrollPage(winUp);
				handled = true;
				break;
			default:
				break;
		}
		break;

	case fldChangedEvent:
		FrmNoteEditUpdateScroll();
		handled = true;
		break;

	case sclRepeatEvent:
		FldScrollField(GetObjectPtr(TaskNote), 
			abs(e->data.sclRepeat.value - e->data.sclRepeat.newValue), 
			e->data.sclRepeat.value > e->data.sclRepeat.newValue ?
			winUp : winDown);
		break;

	default:
		break;
	}

	return handled;
} // static Boolean FrmNoteEditHandleEvent (EventPtr e)



Boolean FrmFlatCtlSelectHandleEvent(EventPtr e)
{
	FormPtr frm;
	
	frm = FrmGetActiveForm();

	switch(e->data.ctlSelect.controlID) {
		case Btn_Done:
			FrmGotoForm(FrmProjectList);
			CloseDB(gdbP);
			gActualTask = gLastSelected = gRefLevel = gParentTask = 0;
			break;
		case Btn_Main_Hierarchical:
			FrmGotoForm(FrmMain);
			break;
		case Btn_Main_Flat:
			FrmPopupForm(FrmFlatFilter);
			break;
		default:
			break;
	}
	FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
	return true;
} // Boolean FrmFlatCtlSelectHandleEvent(EventPtr e)



void ProjectFlatTableUpdate(void)
{
	TablePtr table;
	UInt16 numRows;
	UInt16 i, index;
	UInt16 cardNo = 0;
	LocalID sortInfoID;
	UInt32 *p;
	UInt32 size = 0;

	table = GetObjectPtr(ProjectTable);
	numRows = TblGetNumberOfRows(table);

	// get the sort info block
	DmDatabaseInfo(cardNo, gdbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
		NULL, &sortInfoID, NULL, NULL);

	p = MemLocalIDToLockedPtr(sortInfoID, cardNo);
	size = MemPtrSize(p) / sizeof(UInt32);

	// if gdbP contains at least one task, there is at least one terminal task
	// special case, db empty (just task 0)
	if ((PgNumRecords(gdbP) == 1) || (*p == 0))
	{
		ProjectTableEmpty();
		MemPtrUnlock(p);
		return;
	}

	if (gFlatOffset >= size)
		gFlatOffset = size - 1;

	// first item
	DmFindRecordByID(gdbP, p[gFlatOffset], &index);
	if (index)
	{
		TblSetItemInt(table, 0, 0, index);
		TblSetRowUsable(table, 0, true);
		TblSetRowSelectable(table, 0, true);
	}
	else
	{
		TblSetItemInt(table, 0, 0, -1);
		TblSetRowUsable(table, 0, false);
		TblSetRowSelectable(table, 0, false);
	}

	// first item change if it is now invisible
	// TODO

	// other items
	// TODO : end detection code, etc...
	for (i = 1; i < numRows; i++)
	{
		if (gFlatOffset + i >= size)
		{
			TblSetItemInt(table, i, 0, -1);
			TblSetRowUsable(table, i, false);
		}
		else
		{
			DmFindRecordByID(gdbP, p[gFlatOffset + i], &index);
			if (index)
			{
				TblSetItemInt(table, i, 0, index);
				TblSetRowUsable(table, i, true);
 				TblSetRowSelectable(table, i, true);
			}
			else
			{
				TblSetItemInt(table, i, 0, -1);
				TblSetRowUsable(table, i, false);
 				TblSetRowSelectable(table, i, false);
			}
		}
	}
	MemPtrUnlock(p);
	gCurrentPrefs.topTask = TblGetItemInt(table, 0, 0);
} // void ProjectFlatTableUpdate(void)



Boolean FrmFlatKeyDownHandleEvent(EventPtr e)
{
	switch (e->data.keyDown.chr)
	{
		case vchrPageDown:
			gFlatOffset += 5;
			FrmUpdateForm(FrmFlat, frmRedrawUpdateCode);
			break;
		case vchrPageUp:
			if (gFlatOffset < 5)
				gFlatOffset = 0;
			else
				gFlatOffset -= 5;
			FrmUpdateForm(FrmFlat, frmRedrawUpdateCode);
			break;

		// this code don't work well at all !
		// no time to correct yet
//		case chrUpArrow:
//			if (gActualTask)
//			{
//				UInt16 prev = gActualTask - 1;
//				while (prev && !FlatFilter(gdbP, prev))
//					prev--;
//				if (prev)
//				{
//					gActualTask = prev;
//					SelectActualTask(true);
//					FrmUpdateForm(FrmFlat, frmRedrawUpdateCode);
//				}
//			}
//			break;
//		case chrDownArrow:
//			if ((gActualTask + 1 < DmNumRecords(gdbP)))
//			{
//				UInt16 next = gActualTask + 1, numRec = DmNumRecords(gdbP);
//				while (next < numRec && !FlatFilter(gdbP, next))
//					next++;
//				if (next < numRec)
//				{
//					gActualTask = next;
//					SelectActualTask(true);
//					FrmUpdateForm(FrmFlat, frmRedrawUpdateCode);
//				}
//			}
//			break;
		default:
			break;
	}
	return true;
} // Boolean FrmFlatKeyDownHandleEvent(EventPtr e)



static Boolean FrmFlatHandleEvent (EventPtr e)
{
	Boolean handled = false;
	FormPtr frm;
	Int16 x;
	TaskFormatType format;
	
	frm = FrmGetActiveForm();
	gTable = GetObjectPtr(ProjectTable);

	switch (e->eType) {
	case frmUpdateEvent:
		switch(e->data.frmUpdate.updateCode)
		{
			case frmRedrawUpdateCode:
				ProjectFlatTableUpdate();
				FrmDrawForm(frm);
				SelectActualTask(false);
				handled = true;
				break;
			default:
				break;
		}
		break;

	case frmOpenEvent:
		if (OSCaps.handera)
		{
			VgaFormModify(frm, vgaFormModify160To240);
			VgaTableUseBaseFont(gTable, true);
		}
		gCurrentPrefs.view = flatView;
		ProjectTableInit();
		SetFormTitle(frm, gCurrentPrefs.openDBName);
		FlatCreateList(gdbP);
		ProjectFlatTableUpdate();
		FrmDrawForm(frm);
		gTable = GetObjectPtr(ProjectTable);
		gParentTask = 0;
		gRefLevel = 0;
		CtlSetValue(GetObjectPtr(Btn_Main_Flat), true);
		handled = true;
		break;

	case menuEvent:
		MenuEraseStatus(NULL);
		switch(e->data.menu.itemID) {
			case MenuMainPreferences:
				ShowProjectPropertiesDialog();
				FrmUpdateForm(FrmFlat, frmRedrawUpdateCode);
				break;
			case MenuMainDisplayPreferences:
				FrmPopupForm(FrmDisplayProperties);
				break;
			case MenuMainFlatFilter:
				FrmPopupForm(FrmFlatFilter);
				break;
			case MenuPublishLinkMaster:
				if (gActualTask > 0)
					PublishToLinkMaster(gdbP, gActualTask);
				break;
			case MenuLinkViaLinkMaster:
				if (gActualTask > 0)
					RequestLinkViaLinkMaster(gCurrentPrefs.openDBName, gActualTask);
				break;
			case MenuRemoveLinkViaLinkMaster:
				if (gActualTask > 0)
					RemoveLinkViaLinkMaster(gActualTask);
				break;
			case MenuEditAddNote:
				if (gActualTask)
				{
					FrmPopupForm(FrmNoteEdit);
				}
				break;
			case MenuLinkAll:
				FlatLinkAll(gdbP, true);
				FrmUpdateForm(FrmFlat, frmRedrawUpdateCode);
				break;
			case MenuUnlinkAll:
				FlatLinkAll(gdbP, false);
				FrmUpdateForm(FrmFlat, frmRedrawUpdateCode);
				break;
		}
		handled = true;
		break;

	// remember the screen coords
	case tblEnterEvent:
		gLastDownX = e->screenX;
		gLastDownY = e->screenY;
		// deselect all
		gActualTask = 0;
		break;

	case ctlSelectEvent:
		handled = FrmFlatCtlSelectHandleEvent(e);
		break;

	case ctlRepeatEvent:
		switch (e->data.ctlRepeat.controlID)
		{
			case Btn_Scroll_Up:
				if (gFlatOffset > 0)
					gFlatOffset--;
				break;
			case Btn_Scroll_Down:
					gFlatOffset++;
				break;
			default:
				break;
		}
		FrmUpdateForm(FrmFlat, frmRedrawUpdateCode);
		break;

	case tblSelectEvent:
		gActualTask = TblGetItemInt(gTable, e->data.tblSelect.row, 0);
		if (gActualTask == -1)
			gActualTask = 0;
		format = TaskGetFormat(gdbP, gActualTask);
		x = gLastDownX/11;
		if (gActualTask) // this is for all the switch
		switch (x)
		{
			case 0:
				/**************************************************************
				 * Code for ToDo Link
				 *************************************************************/
				if (format.itemType != extendedType)
				{
					if (format.hasToDo)
						TaskRemoveHasToDo(gdbP, gActualTask);
					else
					{
						TaskSetType(gdbP, gActualTask, actionType);
						TaskSetCompleted(gdbP, gActualTask, ACTION_NOT_DONE,
							true);
						TaskPublishToDo(gdbP, gActualTask);
					}
					FrmUpdateForm(FrmFlat, frmRedrawUpdateCode);
				}
				break;
			case 1:
				/**************************************************************
				 * Code for progress buttons
				 *************************************************************/
				ProgressHandleEvent();
				break;
			default:
				//////////////////////////////////////////////////////////////
				// Code for note
				//////////////////////////////////////////////////////////////
				if (gLastDownX > (OSCaps.screenWidth - 7) && format.hasNote)
				{
					FrmPopupForm(FrmNoteEdit);
				}
				else if (gLastDownX > ((OSCaps.screenWidth - 7) 
					- (format.hasNote? 8 : 0))
					&& format.hasLink)
				{
					FollowLinkViaLinkMaster(gActualTask);
				}
				// if a priority is visible
				else if (x == 2)
				{
					PriorityHandleEvent(e);
				}
				// edit text if click on the right side
				else if (x > 0)
				{
					if (gLastSelected == gActualTask)
					{
						FrmPopupForm(FrmTaskEdit);
					}
					else
					{
						gLastSelected = gActualTask;
					}
				}
				else
				{
					// left part
					// nothing to do for now
				}
				break;
		}
		handled = true;
		break;

	// TODO : put this '5' in prefs (scroll step)
	case keyDownEvent:
		handled = FrmFlatKeyDownHandleEvent(e);
		break;

	// For multi-line items, the table cell is > 1 line high,
    // but PalmOS will not generate a tblSelectEvent if the user taps
	// below the first 11 pixels of the table row. This code manually
	// selects an item no matter where in the cell the tap occurs.
	case penDownEvent:
	{
		Int16 row;

		row = TblGetRowContainingPoint(gTable, e->screenX, e->screenY);
		if(row != -1)
		{
			gActualTask = TblGetItemInt(gTable, row, 0);
			TblSelectItem(gTable, row, 0);
		}
		break;
	}
	default:
		break;
	}

	return handled;
} // static Boolean FrmFlatHandleEvent (EventPtr e)



static Boolean FrmTaskDefaultsHandleEvent (EventPtr e)
{
	Boolean handled = false;
	FormPtr frm;
	ItemType taskType;
	
	frm = FrmGetActiveForm();

	switch (e->eType) {
	case frmOpenEvent:
		if (OSCaps.handera)
		{
			VgaFormModify(frm, vgaFormModify160To240);
		}
		FrmDrawForm(frm);
		// some older releases modified these fields, resulting in initial
		// corrupted notes
		gEmptyTask.note = "";
		gEmptyTask.description = "";
		gProjectPrefs.taskDefaults.note = "";
		gProjectPrefs.taskDefaults.description = "";
	
		// +++ FIX THIS +++
		// How about we figure out how this data structure is getting corrupt in the first place?
		// +++ FIX THIS +++
               // make sure a corrupt data structure does not crash program
	       taskType=gEmptyTask.format.bits.itemType;
               if (taskType >= invalidType)
	       {
			DEBUG1("Something trashed format.bits.itemType, resetting to progressType");
			taskType=progressType;
	       }

		// initialize item type list and popup trigger
		LstSetSelection(GetObjectPtr(Lst_Item_Type), 
			gEmptyTask.format.bits.itemType);

		UpdatePopupTrigger(GetObjectPtr(Pop_Item_Type), 
			GetObjectPtr(Lst_Item_Type));

		// to be compatible with not initialised db
		if (gEmptyTask.priority > 0 && gEmptyTask.priority < 7)
		{
			CtlSetValue(GetObjectPtr(Btn_Priority_X + gEmptyTask.priority - 1),
				true);
		}
		else
		{
			CtlSetValue(GetObjectPtr(Btn_Priority_X + NO_PRIORITY), true);
		}

		handled = true;
		break;
	
	case ctlSelectEvent:
		switch(e->data.ctlSelect.controlID) {
			case Btn_Edit_Ok:
				gProjectPrefs.taskDefaults = gEmptyTask;
				FrmReturnToForm(0);
				handled = true;
				break;

			case Btn_Edit_Cancel:
				gEmptyTask = gProjectPrefs.taskDefaults;
				FrmReturnToForm(0);
				handled = true;
				break;

			case Pop_Item_Type:

				gEmptyTask.format.bits.itemType = 
					LstPopupList(GetObjectPtr(Lst_Item_Type));
				gEmptyTask.completed = 0;
				UpdatePopupTrigger(GetObjectPtr(Pop_Item_Type), 
					GetObjectPtr(Lst_Item_Type));
				break;

			case Btn_Priority_1:
			case Btn_Priority_2:
			case Btn_Priority_3:
			case Btn_Priority_4:
			case Btn_Priority_5:
			case Btn_Priority_No:
				gEmptyTask.priority = e->data.ctlSelect.controlID - 
					Btn_Priority_X + 1;

			default:
				break;
		}
		break;

	default:
		break;
	}

	return handled;
} // static Boolean FrmTaskDefaultsHandleEvent (EventPtr e)



static Boolean ApplicationHandleEvent(EventPtr e)
{
 	if (e->eType == frmLoadEvent)
 	{
 		UInt16  formId = e->data.frmLoad.formID;
 
 		struct tag_HandlerMap {
 			UInt16 id;
 			Boolean (*proc)(EventPtr e);
 		};
 		static struct tag_HandlerMap hm[] = {
 			{ FrmMain,              FrmMainHandleEvent },
 			{ FrmTaskEdit,          FrmTaskEditHandleEvent },
 			{ FrmProjectList,       FrmProjectListHandleEvent },
// 			{ FrmChooseName,        FrmChooseNameHandleEvent },
// 			{ FrmProjectProperties, FrmProjectPropertiesHandleEvent },
 			{ FrmDisplayProperties, FrmDisplayPropertiesHandleEvent },
 			{ FrmNoteEdit,          FrmNoteEditHandleEvent },
 			{ FrmFlat,              FrmFlatHandleEvent },
 			{ FrmFlatFilter,        FrmFlatFilterHandleEvent },
			{ FrmTaskDefaults,		FrmTaskDefaultsHandleEvent },
			{ FrmMemoExportProperties, FrmMemoExportPropertiesHandleEvent },
			{ FrmIconSelect,        FrmIconSelectHandleEvent },
			{ FrmSetReset,			FrmSetResetHandleEvent },
			{ FrmCategoriesSelect,	FrmCategoriesSelectHandleEvent },
 			{ 0, NULL }
 		};
 		static struct tag_HandlerMap* p;
 
 		for (p = hm; p->proc != NULL && p->id != formId; p++)
 			;
 		if (p->proc != NULL)
 		{
 			FormPtr frm = FrmInitForm(formId);
 			FrmSetActiveForm(frm);
 			FrmSetEventHandler(frm, p->proc);
 			return true;
 		}
 	}
   
 	return false;
} // static Boolean ApplicationHandleEvent(EventPtr e)



/* Get preferences, open (or create) app database */
static UInt16 StartApplication(void)
{
	UInt8 i;
	UInt32 version;
	//SonySysFtrSysInfoP infoP;

#if DEBUGMETHOD == 2
	HostTraceInit(); // Start debug output via Palm Reporter
#endif
	//
	// get PalmOS features and settings.
	//

	// Detect PalmOS version to know OS features
	if (! FtrGet(sysFtrCreator, sysFtrNumROMVersion, &version))
	{
#define ver(a,b) sysMakeROMVersion(a, b, 0, sysROMStageRelease, 0)
		OSCaps.ver30 = (version >= ver(3, 0));
		OSCaps.ver31 = (version >= ver(3, 1));
		OSCaps.ver32 = (version >= ver(3, 2));
		OSCaps.ver33 = (version >= ver(3, 3));
		OSCaps.ver34 = (version >= ver(3, 4));
		OSCaps.ver35 = (version >= ver(3, 5));
#undef ver
	}

	// Detect HandEra extensions
//	if (! FtrGet(TRGSysFtrID, TRGVgaFtrNum, &version))
//	{
//		OSCaps.handera = 1;
//		OSCaps.screenWidth  = 240;
//		OSCaps.screenHeight = 240; // with virtual silkscreen on
//		gActualNumRows = 17;
//		VgaSetScreenMode(screenMode1To1, rotateModeNone);
//	}
//	else
//	{
		OSCaps.handera = 0;
		OSCaps.screenWidth  = 160;
		OSCaps.screenHeight = 160;
//	}

	// check for Hires Clie
    // TODO : continue here
    // For now, don't use this or the wrapping won't run correctly
#if 0
	if (!FtrGet(sonySysFtrCreator, sonySysFtrNumSysInfoP, (UInt32*)&infoP))
	{
		if (infoP && (infoP->libr & sonySysFtrSysInfoLibrHR))
		{
			int err = SysLibFind(sonySysLibNameHR, &OSCaps.refNum);
			if (err == sysErrLibNotFound)
			{
				err = SysLibLoad('libr', sonySysFileCHRLib, &OSCaps.refNum);
			}
			if (!err)
			{
				// Hires available
				OSCaps.clieHR = 1;
				OSCaps.screenWidth  = 320;
				OSCaps.screenHeight = 320;
				HROpen(OSCaps.refNum);
			}
		}
	}
#endif

	//
	// retrieve screen color depth, and switch the mode
	//
#if 0
	if (OSCaps.ver30)
	{
		if (OSCaps.ver35)
		{
			WinScreenMode(winScreenModeGetSupportedDepths,
						  NULL, NULL, &OSCaps.nScreenDepth, NULL);
			WinScreenMode(winScreenModeGetSupportsColor,
						  NULL, NULL, NULL, &OSCaps.enableColor);
		}
		else
		{
			FtrGet(sysFtrCreator, sysFtrNumDisplayDepth,
				   &OSCaps.nOldScreenDepth);
			ScrDisplayMode(scrDisplayModeGetSupportedDepths, NULL, NULL,
						   &OSCaps.nScreenDepth, NULL);
			ScrDisplayMode(scrDisplayModeGetSupportsColor, NULL, NULL,
						   NULL, &OSCaps.enableColor);
		}

		if      (OSCaps.nScreenDepth & 0x0080) OSCaps.nScreenDepth = 8;
		else if (OSCaps.nScreenDepth & 0x0008) OSCaps.nScreenDepth = 4;
		else if (OSCaps.nScreenDepth & 0x0002) OSCaps.nScreenDepth = 2;
		else                                   OSCaps.nScreenDepth = 1;

		if (! FtrSet(sysFtrCreator, sysFtrNumDisplayDepth,
					 OSCaps.nScreenDepth))
			ScrDisplayMode(scrDisplayModeSetToDefaults,
						   NULL, NULL, NULL, NULL);
	}
#endif

	// load icon data
	initializeIcon();

	// get date format
	gDateFormat = PrefGetPreference(prefDateFormat);
	gToday = Today();
	gTodayDays = DateToDays(gToday);

	// init the icons handles array
	for (i = 0; i < IconNum; i++)
	{
		gIconHandle[i] = (MemHandle)DmGetResource('Tbmp', BtmOffset + i);
	}

	// read the current prefs
	if (!LoadAppPrefs(false))
	{
		gCurrentPrefs.openDB = false;
		gCurrentPrefs.topTask = -1;
		gActualTask = 0;
		gParentTask = 0;
		gRefLevel = 0;
	}
	else
	{
		gActualTask = gCurrentPrefs.actualTask;
		gParentTask = gCurrentPrefs.parentTask;
		gRefLevel = gCurrentPrefs.refLevel;
	}

	// read the global prefs
	LoadAppPrefs(true);

	return 0;
} // static UInt16 StartApplication(void)



/* Save preferences, close forms, close app database */
static void StopApplication(void)
{
	UInt8 i;
	FrmSaveAllForms();
	FrmCloseAllForms();

	// You have to save your app prefs after all the forms are closed
	// or you will have serious state problems because the forms
	// change the state....
	gCurrentPrefs.actualTask = gActualTask;
	gCurrentPrefs.parentTask = gParentTask;
	gCurrentPrefs.refLevel = gRefLevel;

	SaveAppPrefs(false); // current prefs
	SaveAppPrefs(true);  // saved prefs

	if (gdbP)
	{
		CloseDB(gdbP);
		gdbP = NULL;
	}

	// release the icons handles array
	for (i = 0; i < IconNum; i++)
	{
		DmReleaseResource(gIconHandle[i]);
	}

	// clear the clipboard
	NewClipboard();
	CloseClipboard();

	//
	// restore screen mode
	//
	//FtrSet(sysFtrCreator, sysFtrNumDisplayDepth, OSCaps.nOldScreenDepth);
	//ScrDisplayMode(scrDisplayModeSetToDefaults, NULL, NULL, NULL, NULL);

#if DEBUGMETHOD == 2
	HostTraceClose();  // End debug output via Palm Reporter
#endif

} // static void StopApplication(void)

/* The main event loop */
static void EventLoop(void)
{
	Err err;
	EventType e;

	do {
		EvtGetEvent(&e, evtWaitForever);
		if (! SysHandleEvent (&e))
			if (! MenuHandleEvent (NULL, &e, &err))
				if (! ApplicationHandleEvent (&e))
					FrmDispatchEvent (&e);
	} while (e.eType != appStopEvent);
} // static void EventLoop(void)



void switchView(viewType view)
{
	gCurrentPrefs.view = view;
	switch (view)
	{
	case normalView:
		FrmGotoForm(FrmMain);
		FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
		break;
	case flatView:
		FrmGotoForm(FrmFlat);
		FrmUpdateForm(FrmFlat, frmRedrawUpdateCode);
		break;
	default:
		break;
	}
}


/****************************************************************************
 * Name : GoToItem
 * Desc : open database and select task specified by gotoparams.
 * Parm : 
 * 			-> item info
 * 			-> just launching now?
 * Out  : 
 * Auth : seagull, 22.09.2000 JST
 ***************************************************************************/
static void GoToItem(GoToParamsPtr gotoparams, Boolean launchapp)
{
	UInt16 recordNum = gotoparams->recordNum;
	if (! launchapp)
	{ // app is now on running.
		UInt32 uniqueID;
		if (gdbP)
		{
			UInt16 cardNo;
			LocalID dbID;

			DmOpenDatabaseInfo(gdbP, &dbID, 0, 0, &cardNo, 0);
			if (cardNo == gotoparams->dbCardNo && dbID == gotoparams->dbID)
			{ // require same database
				DmRecordInfo(gdbP, recordNum, NULL, &uniqueID, NULL);
				FrmCloseAllForms();
				DmFindRecordByID(gdbP, uniqueID, &recordNum);
			}
			else
			{ // switch to other database
				FrmCloseAllForms();
				CloseDB(gdbP);
				OpenDBByID(gotoparams->dbCardNo, gotoparams->dbID, &gdbP);
				if (! gdbP)
				{
					FrmGotoForm(FrmProjectList);
					return;
				}
			}
		}
		else
		{
			// We were running, but we didn't have a database open, ie on project list screen
			OpenDBByID(gotoparams->dbCardNo, gotoparams->dbID, &gdbP);
			if (! gdbP)
			{
				FrmGotoForm(FrmProjectList);
				return;
			}
		}
	}

	// OK, the database was opened successflly. select view and record.
	//  NOTE: ummm..I want a view for link published time, not a last selected.
	gLastSelected = gActualTask = recordNum;

	switchView(gCurrentPrefs.view);
} // static void GoToItem(GoToParamsPtr gotoparams, Boolean launchapp)



/* Main entry point; it is unlikely you will need to change this except to
   handle other launch command codes */
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
	Err err;

	if (cmd == sysAppLaunchCmdNormalLaunch ||
		cmd == sysAppLaunchCmdOpenDB ||
		cmd == linkAppLaunchFollowLink ||
		cmd == linkAppLaunchReturnLink ||
		cmd == linkAppLaunchRequestLink )
	{
   		err = StartApplication();
		if (err) return err;

		//
		// Open database (,and goto item)
		//
		if (cmd == sysAppLaunchCmdNormalLaunch ||
			cmd == linkAppLaunchRequestLink)
		{
			if (gCurrentPrefs.openDB && ! OpenDB(gCardNo,gCurrentPrefs.openDBName, &gdbP))
				switchView(gCurrentPrefs.view);
			else
				FrmGotoForm(FrmProjectList);
		}
		else if (cmd == sysAppLaunchCmdOpenDB)
		{
			SysAppLaunchCmdOpenDBType* db = (SysAppLaunchCmdOpenDBType*)cmdPBP;
			err = OpenDBByID(db->cardNo, db->dbID, &gdbP);
			if (err) return err;  // Failed to open that database.
			switchView(gCurrentPrefs.view);
		}
		else if (cmd == linkAppLaunchFollowLink)
		{
			GoToParamsPtr gotoparams = LinkSimpleToGoToParams((LinkSimplePtr)cmdPBP);
			if (gotoparams)
			{
				err = OpenDBByID(gotoparams->dbCardNo, gotoparams->dbID, &gdbP);
				if (!err)
					GoToItem(gotoparams, false);
				MemPtrFree(gotoparams);
			}
			else return 0;
		}
		else if (cmd == linkAppLaunchReturnLink)
		{
			LinkSimplePtr p1 = (LinkSimplePtr)cmdPBP;
			LinkInfoPtr p2 = (LinkInfoPtr)((LinkInfoPtr)cmdPBP + 1);
			if (! OpenDB(gCardNo,p1->db_name, &gdbP))
			{
				SetLinkMasterLink(p1, p2);
				DmFindRecordByID(gdbP, p1->record_id, &gActualTask);
				switchView(gCurrentPrefs.view);
			}
			else
				FrmGotoForm(FrmProjectList);
		}

		EventLoop();
		StopApplication();
	}
	else if (cmd == linkAppLaunchFollowLinkSubCall)
	{
		GoToParamsPtr gotoparams;
		if (launchFlags & sysAppLaunchFlagNewGlobals)
		{
			gotoparams = LinkSimpleToGoToParams((LinkSimplePtr)cmdPBP);
			if (gotoparams)
			{
				GoToItem(gotoparams, false);
				MemPtrFree(gotoparams);
			}
		}
		else
		{
			DmSearchStateType state;
			UInt16 card;
			LocalID db;

			if (DmGetNextDatabaseByTypeCreator(true, &state, sysFileTApplication,
											   CREATOR, false, &card, &db) == 0)
				SysUIAppSwitch(card, db, linkAppLaunchFollowLink, cmdPBP);
		}
	}
	else if (cmd == sysAppLaunchCmdFind)
	{
		Search ((FindParamsPtr)cmdPBP);
		// open first database
		// if it doesn't exist
		// 		findParams->more = false;
		// 		exit
		// while we've got a database
		// 		scan each record
		// 		search for the next database
		// exit
		return 0;
	}
	else if (cmd == sysAppLaunchCmdGoTo)
	{
		GoToParamsPtr gotoparams = (GoToParamsPtr)cmdPBP;
		Boolean launched = launchFlags & sysAppLaunchFlagNewGlobals;
		if (launched)
		{
			err = StartApplication();
			if (err) return err;
			err = OpenDBByID(gotoparams->dbCardNo, gotoparams->dbID, &gdbP);
			if (! err)
			{
				GoToItem(gotoparams, launched);
				EventLoop();
				StopApplication();
			}
		}
		else GoToItem(gotoparams, launched);
	}
	else if (cmd == sysAppLaunchCmdExgAskUser)
		// The reception dialog is skipped.
	{
		ExgAskParamPtr paramP = (ExgAskParamPtr) cmdPBP;
		paramP->result = exgAskOk; // exgAskDialog;
		return 0;
	}
	else if (cmd == sysAppLaunchCmdExgReceiveData)
	{
		Err err = RecvDB((ExgSocketPtr) cmdPBP);
		// FIXME? if progect is not running...
		FrmUpdateForm(FrmProjectList, frmRedrawUpdateCode);
		return err;
	}		
//	else if (cmd == sysAppLaunchCmdSyncNotify)
//		// Any received file whose name extension is ".pgt" is for us
//		return ExgRegisterData(CREATOR, exgRegExtensionID, EXTENSION);
	else if (cmd == sysAppLaunchCmdSyncNotify)
	{
		// TODO : Launch aftersync;
		afterSync();
	}
	else
		return sysErrParamErr;

	return 0;
} // UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)



/*
 * based on MemoMain.c from Palm
 * adapted by L. Burgbacher for Progect
 */
static void DrawProjectTitle (Char * memo, Int16 x, Int16 y, Int16 width)
{
	Char * ptr = StrChr (memo, linefeedChr);
	UInt16 titleLen = (ptr == NULL ? StrLen (memo) : (UInt16) (ptr - memo));
	if (FntWidthToOffset (memo, titleLen, width, NULL, NULL) == titleLen)
		{
		WinDrawChars (memo, titleLen, x, y);
		}
	else
		{
		Int16 titleWidth;
		titleLen = FntWidthToOffset (memo, titleLen, width - FntCharWidth (chrEllipsis), NULL, &titleWidth);
		WinDrawChars (memo, titleLen, x, y);
		WinDrawChar (chrEllipsis, x + titleWidth, y);
		}
}



/*
 * based on MemoMain.c from Palm
 * adapted by L. Burgbacher for Progect
 */
static void Search(FindParamsPtr findParams)
{
	UInt16 pos;
	Char * header;
	UInt16 recordNum;
	MemHandle h;
	MemHandle headerStringH;
	RectangleType r;
	Boolean done;
	Boolean match;
	DmOpenRef dbP;
	DmSearchStateType searchState;
	Err err;
	UInt16 cardNo = 0;
	LocalID dbID, currentDbID;
	TaskRecordType *p;
	UInt32 longPos;
	UInt16 matchLength;
	Char searchDbName[] = "lbPG-GlobalSearch";
	LocalID searchID = 0;
	LocalID *IDPtr = NULL;
	DmOpenRef searchDb = NULL;
	Boolean firstTime = true;

	// if this is a continuation, find the last opened database
	if (findParams->continuation)
	{
		searchID = DmFindDatabase(cardNo, searchDbName);

		if (!searchID)
		{
			// this denotes a problem...
			findParams->more = false;
			return;
		}

		searchDb = DmOpenDatabase(cardNo, searchID, dmModeReadWrite);

		if (!searchDb)
		{
			// this denotes a problem...
			findParams->more = false;
			return;
		}

		h = DmQueryRecord(searchDb, 0);
		if (h)
		{
			currentDbID = *(LocalID*)(MemHandleLock(h));
			MemHandleUnlock(h);
			DmCloseDatabase(searchDb);
		}
		else
		{
			findParams->more = false;
			DmCloseDatabase(searchDb);
			DmDeleteDatabase(cardNo, searchID);
			return;
		}

	}
	else
	{
		// init currentDbID for first time
		DmGetNextDatabaseByTypeCreator (true, &searchState, 'DATA',
					CREATOR, false, &cardNo, &currentDbID);
	}
	
	// find the actual searched db
	while (dbID != currentDbID)
	{
		// Find the application's data file.
		err = DmGetNextDatabaseByTypeCreator (firstTime, &searchState, 'DATA',
					CREATOR, false, &cardNo, &dbID);
		firstTime = false;
		if (err)
		{
			findParams->more = false;
			if (searchID)
				DmDeleteDatabase(cardNo, searchID);
			return;
		}
	}

   // Open the first database.
   dbP = DmOpenDatabase(cardNo, dbID, findParams->dbAccesMode);
   if (!dbP)
   	{
    	findParams->more = false;
		if (searchID)
			DmDeleteDatabase(cardNo, searchID);
      	return;
   	}

	// Display the heading line.
	headerStringH = DmGetResource('tSTR', StrFindHeader);
	header = MemHandleLock(headerStringH);
	done = FindDrawHeader(findParams, header);
   	MemHandleUnlock(headerStringH);   
   	DmReleaseResource(headerStringH);   
   	if (done) 
      goto Exit;

	// Search the memos for the "find" string.
	recordNum = findParams->recordNum;
	while (true)
		{
		// Because applications can take a long time to finish a find when
		// the result may be on the screen or for other reasons, users like
		// to be able to stop the find.  Stop the find if an event is pending.
		// This stops if the user does something with the device.  Because
		// this call slows winDown the search we perform it every so many 
		// records instead of every record.  The response time should still
		// be Int16 without introducing much extra work to the search.
		if ((recordNum & 0x000f) == 0 &&			// every 16th record
			EvtSysEventAvail(true))
			{
			// Stop the search process.
			findParams->more = true;
			break;
			}

		h = DmQueryRecord(dbP, recordNum);

		// Have we run out of records?
		while (!h)
		{
			// search the next database
			err = DmGetNextDatabaseByTypeCreator (false, &searchState, 'DATA',
						CREATOR, false, &cardNo, &dbID);
			if (err)
			{
				findParams->more = false;
				if (searchID)
					DmDeleteDatabase(cardNo, searchID);
				break;
			}
		   	DmCloseDatabase(dbP);

		   dbP = DmOpenDatabase(cardNo, dbID, findParams->dbAccesMode);
		   if (!dbP)
			{
			  findParams->more = false;
				if (searchID)
					DmDeleteDatabase(cardNo, searchID);
			  break;
			}

			recordNum = 1;
			findParams->recordNum = 1;
			h = DmQueryRecord(dbP, recordNum);
		}

		if (!h) break;

		p = MemHandleLock(h);
		 
		// don't treat extended types yet.
		if (p->format.bits.itemType == extendedType)
		{
			recordNum++;
			MemHandleUnlock(h);
			continue;
		}

		// Search for the string passed,  if it's found display the title
		// of the project.
//		match = TxtGlueFindString (&(p->description), findParams->strToFind, &longPos, &matchLength);
		match = TxtFindString(&(p->description), findParams->strToFind, &longPos, &matchLength);
		if (!match)
		{
		char *note;
		/*
			note = TaskGetNote(dbP, recordNum);
			if (note)
				match = TxtFindString(note, findParams->strToFind, &longPos, &matchLength);
		*/
		
			note = &(p->description);
			note += StrLen(note) + 1;
			if (note <= (Char*)p + MemHandleSize(h))
			{
				match = TxtFindString(note, findParams->strToFind, &longPos, &matchLength);
			}

		}
		pos = longPos;
		if (match)
			{
			// Add the match to the find paramter block,  if there is no room to
			// display the match the following function will return true.
			done = FindSaveMatch (findParams, recordNum, pos, 0, matchLength, cardNo, dbID);

			if (!done)
				{
				// Get the bounds of the region where we will draw the results.
				FindGetLineBounds (findParams, &r);
				
				// Display the title of the description.
				DrawProjectTitle (&(p->description), r.topLeft.x+1, r.topLeft.y, 
					r.extent.x-2);
	
				findParams->lineNumber++;
				}

			}
		
		MemHandleUnlock(h);
		if (done) break;

		recordNum++;
		}

Exit:
	if (dbP)
		DmCloseDatabase (dbP);	
	if (findParams->more)
	{
		// create the searchdatabase and init it
		err = DmCreateDatabase(cardNo, searchDbName, CREATOR, 'SRCH', false);
		if ((err == dmErrAlreadyExists) || !err)
		{
			UInt16 index = 0;
			searchID = DmFindDatabase(cardNo, searchDbName);

			if (!searchID) return;

			searchDb = DmOpenDatabase(cardNo, searchID, dmModeReadWrite);

			if (!searchDb) return;

			while (DmNumRecords(searchDb))
			{
				DmRemoveRecord(searchDb, 0);
			}

			h = DmNewRecord(searchDb, &index, sizeof(LocalID));
			if (h)
				IDPtr = MemHandleLock(h);
			if (IDPtr)
			{
				DmWrite(IDPtr, 0, &dbID, sizeof(LocalID));
				MemHandleUnlock(h);
				DmReleaseRecord(searchDb, index, true);
			}
			DmCloseDatabase(searchDb);
		}
	}
}

/****************************************************************************
 * Name : ViewCreateNewTask
 * Desc : Create new task
 * Parm : 
 * Out  :
 * Auth : rick_price
 * Mod  : 
 ***************************************************************************/
void ViewCreateNewTask(void)
{
	if (gActualTask ||
		PgNumRecords(gdbP) == 1 ||
		!TaskGetHasChild(gdbP, gParentTask))
		{
			// special case when there's no existing task
			if (PgNumRecords(gdbP) == 1 || 
				!TaskGetHasChild(gdbP, gParentTask))
			{
				gActualTask = gParentTask;
			}
			AddRecord(gdbP, &gEmptyTask, &gActualTask);

			// recalc fathers progressbar
			if (TaskGetLevel(gdbP, gActualTask) > 1)
			{
				TaskCalcCompleted(gdbP,
						  TaskGetFatherIndex(gdbP, gActualTask));
			}
			FrmPopupForm(FrmTaskEdit);
			SelectActualTask(true);
			FrmUpdateForm(FrmMain, frmRedrawUpdateCode);
		}
}

